import asyncio
from dataclasses import dataclass
from functools import partial
import bspump
import os
from typing import Any, Callable, List

# Define globals if not define already
if "bitswan_auto_pipeline" not in globals():
    __bitswan_processors = []
    __bitswan_pipelines = {}
    __bitswan_current_pipeline = None
    __bitswan_dev_runtime = None
    __bitswan_connections = []
    __bitswan_lookups = []
    _bitswan_app_post_inits = []
    bitswan_auto_pipeline = {}
    __bs_step_locals = {}
    bitswan_test_mode = []
    __bitswan_autopipeline_count = 1
    bitswan_test_probes = {}
    bitswan_tested_pipelines = set()


class DevPipeline(bspump.Pipeline):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DevApp(bspump.BSPumpApplication):
    def __init__(self, args=[]):
        super().__init__(args=args)


class DevRuntime:
    def __init__(self, args=[]):
        self.events: list[tuple[str, list[Any]]] = []
        self.dev_app = DevApp(args)

    def clear(self, name: str, event: list[Any]) -> None:
        self.events = [(name, event)]

    def get_prev_events(self, name) -> tuple[list[tuple[str, list[Any]]], list[Any]]:
        new_eventss = []
        prev_events = []
        for event in self.events:
            if event[0] == name:
                break
            prev_events = event[1]
            new_eventss.append(event)
        return new_eventss, prev_events

    def step(self, name: str, func) -> None:
        new_eventss, prev_events = self.get_prev_events(name)
        new_events = [func(prev_event) for prev_event in prev_events]
        [print(event) for event in new_events]
        new_eventss.append((name, new_events))
        self.events = new_eventss

    async def async_step(self, name: str, func) -> None:
        new_eventss, prev_events = self.get_prev_events(name)
        new_events = []

        async def inject(event):
            new_events.append(event)

        [await func(inject, prev_event) for prev_event in prev_events]
        [print(event) for event in new_events]
        new_eventss.append((name, new_events))
        self.events = new_eventss


def auto_pipeline(source=None, sink=None, name=None):
    if source is None:
        raise Exception(
            "When calling auto_pipeline must specify a function that returns a source."
        )
    if sink is None:
        raise Exception(
            "When calling auto_pipeline must specify a function that returns a sink."
        )

    global __bitswan_autopipeline_count
    if name is None:
        pipeline_name = f"auto_pipeline_{__bitswan_autopipeline_count}"
    else:
        pipeline_name = f"{name}"
    new_pipeline(pipeline_name)
    __bitswan_autopipeline_count += 1
    import inspect

    frame = inspect.currentframe()
    register_source(source, test_events=frame.f_back.f_locals.get("test_events"))
    bitswan_auto_pipeline["sink"] = sink


async def test_devruntime():
    test_runtime = DevRuntime()

    def test_func(prev_event):
        return prev_event + 1

    tr = test_runtime
    tr.clear("sample", [1, 2, 3])
    tr.step("step1", test_func)
    tr.step("step2", test_func)
    tr.step("step3", test_func)
    assert tr.events == [
        ("sample", [1, 2, 3]),
        ("step1", [2, 3, 4]),
        ("step2", [3, 4, 5]),
        ("step3", [4, 5, 6]),
    ]

    def test_func1(prev_event):
        return prev_event + 7

    tr.step("step2", test_func1)
    assert tr.events == [
        ("sample", [1, 2, 3]),
        ("step1", [2, 3, 4]),
        ("step2", [9, 10, 11]),
    ]

    async def test_func2(injector, prev_event):
        if prev_event % 2 == 0:
            injector(prev_event + 1)

    await tr.async_step("step3", test_func2)
    assert tr.events == [
        ("sample", [1, 2, 3]),
        ("step1", [2, 3, 4]),
        ("step2", [9, 10, 11]),
        ("step3", [11]),
    ]


def is_running_in_jupyter():
    try:
        from IPython import get_ipython

        # Check if the IPython kernel is running which is a strong indication of a Jupyter environment
        if "IPKernelApp" in get_ipython().config:
            return True
        if (
            "VSCODE_PID" in os.environ
        ):  # Check for Visual Studio Code's Jupyter extension
            return True
        return False
    except Exception:
        return False


__bitswan_dev = is_running_in_jupyter()


def ensure_bitswan_runtime(func):
    def wrapper(*args, **kwargs):
        global __bitswan_dev_runtime
        if not __bitswan_dev_runtime:
            __bitswan_dev_runtime = DevRuntime()

        return func(*args, **kwargs)

    return wrapper


def init_bitswan_jupyter(config_path: str = None):
    """Helper function to initialize Bitswan in Jupyter environment,
    needs to be called before any other Bitswan function.
    Does not need to be called if config file is not needed

    Args:
        config_path (str, optional): Path to the config file. Defaults to None.
    """
    global __bitswan_dev_runtime

    args = ["-c", config_path] if config_path else []
    __bitswan_dev_runtime = DevRuntime(args)


def add_test_probe(name):
    global bitswan_test_mode
    global bitswan_test_probes
    if bitswan_test_mode:
        import inspect

        frame = inspect.currentframe()
        try:
            probe, expected = bitswan_test_probes.get(name, (None, None))
            if probe is not None:
                print(f"    │ Probing {name}.")
                probed = probe(frame.f_back.f_locals, frame.f_back.f_globals)
                if not probed == expected:
                    print(
                        "    └ \033[91m"
                        + f"Probe {name} failed. Got {probed} expected {expected}."
                        + "\033[0m"
                    )
                    exit(1)
        finally:
            del frame


@ensure_bitswan_runtime
def sample_events(events: List[Any]):
    """Inject sample events into the current pipeline for testing

    Args:
        events (List[Any]): List of events to be injected
    """
    global __bitswan_dev_runtime
    __bitswan_dev_runtime.clear("__sample", events)


def register_app_post_init(func):
    """
    Ex:
    @register_app_post_init
    def post_init(app):
        app.PubSub.subscribe("Application.tick!", app.tick)
    """
    global _bitswan_app_post_inits
    _bitswan_app_post_inits.append(func)


@ensure_bitswan_runtime
def register_connection(func):
    """
    Ex:
    @register_connection
    def connection(app):
        return bspump.kafka.KafkaConnection(app, "KafkaConnection")
    """
    global __bitswan_connections
    global __bitswan_dev_runtime
    __bitswan_connections.append(func)
    if __bitswan_dev:
        global __bitswan_dev_runtime
        connection = func(__bitswan_dev_runtime.dev_app)
        __bitswan_dev_runtime.dev_app.PumpService.add_connection(connection)


async def retrieve_sample_events(limit: int = 10) -> None:
    """Get sample events from the source registered to the current pipeline and register them for testing,
    has to be awaited in Jupyter environment

    Ex:
        await retrieve_sample_events(100)

    Args:
        limit (int, optional): Number of events to retrieve. Defaults to 10.

    Returns:
        None


    """
    global __bitswan_dev_runtime
    # Capture the current state of __bitswan_processors
    current_processors = list(__bitswan_processors)

    class TmpPipeline(bspump.Pipeline):
        def __init__(self, app, pipeline_id):
            super().__init__(app, pipeline_id)
            processors = []
            for processor in current_processors:
                instance = processor(app, self)
                processors.append(instance)
            self.build(*processors)
            self.events = []

        def inject(self, context, event, depth):
            if len(self.events) >= limit:
                return
            print(event)
            self.events.append(event)

        def get_events(self):
            print(f"Collected {len(self.events)} events")
            return self.events

    pipeline = TmpPipeline(__bitswan_dev_runtime.dev_app, __bitswan_current_pipeline)
    pipeline.start()

    try:
        while 1:
            if len(pipeline.events) >= limit:
                await pipeline.stop()
                break
            await asyncio.sleep(0.5)
    except asyncio.CancelledError:
        await pipeline.stop()

    __bitswan_dev_runtime.clear("__sample", pipeline.get_events())
    return


@ensure_bitswan_runtime
def register_lookup(func):
    """
    Ex:
    @register_lookup
    def lookup(app):
      return ENodeBLookup(app, id='ENodeBLookup')
    """
    global __bitswan_lookups
    __bitswan_lookups.append(func)
    if __bitswan_dev:
        global __bitswan_dev_runtime
        lookup = func(__bitswan_dev_runtime.dev_app)
        __bitswan_dev_runtime.dev_app.PumpService.add_lookup(lookup)


def new_pipeline(name: str):
    """Creates and registers a new pipeline

    Args:
        name (str): Name of the pipeline
    """
    global __bitswan_processors
    global __bitswan_current_pipeline
    __bitswan_processors = []
    __bitswan_current_pipeline = name


def end_pipeline():
    """Ends the current pipeline and appends it to the list of pipelines"""
    global __bitswan_current_pipeline
    global __bitswan_processors
    global __bitswan_pipelines
    global __bitswan_dev

    if __bitswan_dev:
        return

    # Capture the current state of __bitswan_processors
    current_processors = list(__bitswan_processors)

    class Pipeline(bspump.Pipeline):
        def __init__(self, app, pipeline_id):
            super().__init__(app, pipeline_id)
            processors = []
            for processor in current_processors:
                instance = processor(app, self)
                processors.append(instance)
            self.build(*processors)

    # Append the new Pipeline class
    __bitswan_pipelines[__bitswan_current_pipeline] = Pipeline


def register_source(func, test_events=None):
    """
    Ex:
    @register_source
    def source(app, pipeline):
        return bspump.socket.TCPStreamSource(app, pipeline)
    """
    if test_events is None:
        import inspect

        frame = inspect.currentframe()
        test_events = frame.f_back.f_locals.get("test_events", {})
    global __bitswan_processors
    global bitswan_test_mode
    if bitswan_test_mode:
        import bspump.test

        def test_source(app, pipeline):
            return bspump.test.TestSource(app, pipeline, test_events=test_events)

        __bitswan_processors.append(test_source)
    else:
        __bitswan_processors.append(func)


@ensure_bitswan_runtime
def register_processor(func):
    """
    Ex:
    @register_processor
    def processor(app, pipeline):
        return bspump.socket.TCPStreamProcessor(app, pipeline)
    """
    global __bitswan_processors
    global __bitswan_dev
    global __bitswan_dev_runtime
    global __bitswan_current_pipeline
    if not __bitswan_dev:
        __bitswan_processors.append(func)
    else:
        pipeline = DevPipeline(
            app=__bitswan_dev_runtime.dev_app, id=__bitswan_current_pipeline
        )
        processor = func(__bitswan_dev_runtime.dev_app, pipeline)
        callable_process = partial(processor.process, processor)
        __bitswan_dev_runtime.step(func.__name__, callable_process)


@ensure_bitswan_runtime
def register_generator(func):
    """
    Ex:
    @register_generator
    def generator(app, pipeline):
        return MyGeneratorClass(app, pipeline)
    """

    async def _fn():
        global __bitswan_processors
        global __bitswan_dev
        global __bitswan_dev_runtime
        global __bitswan_current_pipeline
        if not __bitswan_dev:
            # TODO: check this
            __bitswan_processors.append(func)
        else:
            pipeline = DevPipeline(
                __bitswan_dev_runtime.dev_app, __bitswan_current_pipeline
            )
            generator = func(__bitswan_dev_runtime.dev_app, pipeline)

            async def asfunc(inject, event):
                async def super_inject(context, event, depth):
                    return await inject(event)

                pipeline.inject = super_inject
                return await generator.generate(None, event, 0)

            await __bitswan_dev_runtime.async_step(func.__name__, asfunc)

        return func

    def wrapper(*args, **kwargs):
        import asyncio
        import nest_asyncio

        nest_asyncio.apply()
        loop = asyncio.get_event_loop()
        task = asyncio.ensure_future(_fn())
        loop.run_until_complete(task)

    wrapper()
    return _fn


def register_sink(func):
    """
    Ex:
    @register_sink
    def sink(app, pipeline):
        return bspump.socket.TCPStreamSink(app, pipeline)
    """
    global __bitswan_processors
    global bitswan_test_mode
    if bitswan_test_mode:
        import bspump.test

        def test_sink(app, pipeline):
            return bspump.test.TestSink(app, pipeline)

        __bitswan_processors.append(test_sink)
    else:
        __bitswan_processors.append(func)


def snake_to_camel_case(name):
    """Convert snake_case name to CamelCase."""
    return "".join(word.capitalize() for word in name.split("_"))


@ensure_bitswan_runtime
def step(func: Callable[[Any], Any]):
    """Decorator that registers a new processor with the given function

    Ex:
        @step
        def my_processor(event):
            return {"processed": event}

    Args:
        func (Callable[[Any], Any]): Function to be registered as a processor
    """
    global __bitswan_processors
    global __bitswan_dev
    global __bitswan_dev_runtime
    if not __bitswan_dev:
        # Convert function name from snake case to CamelCase and create a unique class name
        class_name = snake_to_camel_case(func.__name__) + "Processor"

        # Dynamically create a new Processor class with the custom class name
        CustomProcessor = type(
            class_name,
            (bspump.Processor,),
            {"process": lambda self, context, event: func(event)},
        )

        # Append the new Processor to the __bitswan_processors list
        __bitswan_processors.append(CustomProcessor)
    else:
        __bitswan_dev_runtime.step(func.__name__, func)

    # Return the original function unmodified
    return func


def async_step(func):
    # Convert function name from snake case to CamelCase and create a unique class name
    global __bitswan_dev
    class_name = snake_to_camel_case(func.__name__) + "Generator"

    # Dynamically create a new Generator class with the custom class name
    async def _generate(self, context, event, depth):
        async def injector(event):
            # TODO: do better
            if __bitswan_dev:
                return await self.Pipeline.inject(context, event, depth)
            else:
                return self.Pipeline.inject(context, event, depth)

        return await func(injector, event)

    CustomGenerator = type(
        class_name,
        (bspump.Generator,),
        # Async generate function calls func with injector and event. The injector is taken from the pipeline.
        {"generate": _generate},
    )

    def generator(app, pipeline):
        return CustomGenerator(app, pipeline)

    generator.__name__ = func.__name__
    register_generator(generator)


def _init_pipelines(app, service):
    global __bitswan_pipelines
    for name, pipeline in __bitswan_pipelines.items():
        service.add_pipeline(pipeline(app, name))


def _init_connections(app, service):
    global __bitswan_connections
    for connection in __bitswan_connections:
        service.add_connection(connection(app))


def _init_lookups(app, service):
    global __bitswan_lookups
    for lookup in __bitswan_lookups:
        service.add_lookup(lookup(app))


class App(bspump.BSPumpApplication):
    def init_componets(self):
        global _bitswan_app_post_inits
        svc = self.get_service("bspump.PumpService")
        _init_connections(self, svc)
        _init_lookups(self, svc)
        _init_pipelines(self, svc)
        for func in _bitswan_app_post_inits:
            func(self)


def deploy():
    import os
    import json

    deploy_secret = os.environ.get("BITSWAN_DEPLOY_SECRET", None)

    is_vscode = "VSCODE_PID" in os.environ

    @dataclass
    class DeployDetails:
        notebook_json: dict[str, Any]
        deploy_secret: str
        deploy_url: str

    def get_deploy_details() -> DeployDetails | None:
        """Get the notebook JSON, deploy secret and deploy URL

        Raises:
            e: RuntimeError: No VSCode or Colab environment detected. Please run this in a different environment.

        Returns:
            DeployDetails | None: DeployDetails object containing the notebook JSON, deploy secret and deploy URL
        """

        if is_vscode:
            from IPython import get_ipython

            ip = get_ipython()
            if not ip:
                return None

            notebook_path = os.path.abspath(ip.user_ns.get("__vsc_ipynb_file__", ""))

            if not notebook_path or not notebook_path.endswith(".ipynb"):
                return None

            with open(notebook_path, "r", encoding="utf-8") as f:
                notebook_json = json.load(f)

            details = DeployDetails(
                json.dumps(notebook_json, sort_keys=True, indent=2),
                os.environ["BITSWAN_DEPLOY_SECRET"],
                os.environ[
                    "BITSWAN_DEPLOY_URL"
                ],  # raises KeyError if not set, to match google colab behavior
            )
            return details
        else:
            try:
                from google.colab import _message, userdata
            except ImportError:
                raise RuntimeError(
                    "No VSCode or Colab environment detected. Please run this in a different environment."
                )

            notebook_json_string = _message.blocking_request(
                "get_ipynb", request="", timeout_sec=None
            )
            details = DeployDetails(
                json.dumps(notebook_json_string["ipynb"], sort_keys=True, indent=2),
                userdata.get("BITSWAN_DEPLOY_SECRET"),
                userdata.get("BITSWAN_DEPLOY_URL"),
            )
            return details

    deploy_details = get_deploy_details()
    if not deploy_details:
        print("Error retrieving notebook contents")
        return
    deploy_url = os.path.join(
        deploy_details.deploy_url,
        "__jupyter-deploy-pipeline/",
        "?secret=" + deploy_secret + "&restart=true",
    )
    print("Packing for deployment...")
    import zipfile

    with zipfile.ZipFile("main.zip", "w") as myzip:
        myzip.writestr("main.ipynb", deploy_details.notebook_json)
    import requests

    print("Uploading to server..")
    with open("main.zip", "rb") as f:
        r = requests.post(deploy_url, data=f)
        print(json.loads(r.text)["status"])



from aiohttp.web import Request

from bspump.http.web.components.base_field import BaseField
from bspump.http.web.template_env import env


class Field(BaseField):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        if "___" in name:
            raise ValueError("Field name cannot contain '___'")
        su = super()
        su.__init__(name, **kwargs)
        self.readonly: bool = kwargs.get("readonly", False)
        self.default = kwargs.get("default", "")
        self.field_name: str = f"f___{self.name}"
        self.default_classes = kwargs.get(
            "default_css_classes",
            "bg-slate-50 autofill:shadow-[inset_0_0_0px_1000px_theme(colors.slate.50)] autofill:bg-slate-50 border-2 border-slate-200 text-gray-900 text-md rounded-lg shadow-lg shadow-slate-100 focus:outline-none focus:ring-slate-400 focus:border-slate-400 focus:bg-slate-50 block w-full p-2.5 transition-shadow duration-200 ease-in-out autofill:bg-slate-50",
        )
        if self.readonly:
            self.default_classes = kwargs.get(
                "default_css_classes",
                "bg-slate-100 border-2 border-slate-200 text-gray-900 text-md rounded-lg shadow-lg shadow-slate-100 focus:outline-none focus:ring-none block w-full p-2.5",
            )

    @property
    def default_input_props(self):
        if self.readonly:
            readonly = "readonly"
        else:
            readonly = ""

        if self.required:
            required = 'required aria-required="true"'
        else:
            required = ""
        return f'name="{self.field_name}" id="{self.field_name}" {readonly} {required}'

    def restructure_data(self, dfrom, dto):
        dto[self.name] = dfrom.get(self.field_name, self.default)

    def clean(self, data, request: Request | None = None):
        pass

    def html(self, default=""):
        if not default:
            default = self.default

        template = env.get_template("field.html")
        return template.render(
            field_name=self.field_name,
            display=self.display,
            default_classes=self.default_classes,
            hidden=self.hidden,
            inner_html=self.inner_html(default, self.readonly),
        )

    def get_params(self) -> dict:
        return {"type": type(self).__name__, "description": self.description}



class WebRequestsMetrics(object):
    def __init__(self, metrics_svc):
        self.MetricsService = metrics_svc

        self.MaxDurationCounter = self.MetricsService.create_aggregation_counter(
            "web_requests_duration_max",
            help="Counts maximum request duration to asab endpoints per minute.",
            unit="seconds",
            aggregator=max,
            dynamic_tags=True,
        )
        self.MinDurationCounter = self.MetricsService.create_aggregation_counter(
            "web_requests_duration_min",
            help="Counts minimal request duration to asab endpoints per minute.",
            unit="seconds",
            aggregator=min,
            dynamic_tags=True,
        )
        self.DurationCounter = self.MetricsService.create_counter(
            "web_requests_duration",
            unit="seconds",
            help="Counts total requests duration to asab endpoints per minute.",
            dynamic_tags=True,
        )
        self.RequestCounter = self.MetricsService.create_counter(
            "web_requests",
            unit="epm",
            help="Counts requests to asab endpoints as events per minute.",
            dynamic_tags=True,
        )
        self.DurationHistogram = self.MetricsService.create_histogram(
            "web_requests_duration_hist",
            buckets=[0.0001, 0.0005, 0.001, 0.005, 0.01, 0.05, 0.1, 1, 5, 10, 50],
            unit="seconds",
            help="Categorizes requests based on their duration.",
            dynamic_tags=True,
        )

    def set_metrics(self, duration, method, path, status):
        tags = {"method": method, "path": path, "status": str(status)}

        # max
        self.MaxDurationCounter.set("duration", duration, tags=tags)
        # min
        self.MinDurationCounter.set("duration", duration, tags=tags)
        # count
        self.RequestCounter.add("count", 1, tags=tags)
        # total duration
        self.DurationCounter.add("duration", duration, tags=tags)
        # counts in buckets
        self.DurationHistogram.set("duration", duration, tags=tags)



from unittest.mock import MagicMock

import bspump.unittest
import bspump.common


class TestDictToJsonParser(bspump.unittest.ProcessorTestCase):
    def test_dict_to_json(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
            (None, {"spam": "eggs"}),
        ]

        self.set_up_processor(bspump.common.DictToJsonParser)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output],
            ['{"foo": "bar"}', '{"fizz": "buzz"}', '{"spam": "eggs"}'],
        )

    def test_event_not_dict(self):
        events = [
            "Not a dictionary",
        ]
        bspump.pipeline.L = MagicMock()  # turn off logging
        self.set_up_processor(bspump.common.DictToJsonParser)

        output = self.execute(events)

        self.assertTrue(self.Pipeline.is_error())
        self.assertEqual([event for context, event in output], [])


class TestJsonToDictParser(bspump.unittest.ProcessorTestCase):
    def test_json_to_dict(self):
        events = {
            (None, '{"key": "1"}'),
            (None, '{"key": "2"}'),
            (None, '{"key": "3"}'),
        }

        self.set_up_processor(bspump.common.StdJsonToDictParser)

        output = self.execute(events)

        self.assertListEqual(
            sorted([event for context, event in output], key=lambda d: int(d["key"])),
            [{"key": "1"}, {"key": "2"}, {"key": "3"}],
        )

    def test_event_not_str(self):
        events = [
            (None, {"Not a": "string"}),
        ]
        self.set_up_processor(bspump.common.StdJsonToDictParser)

        output = self.execute(events)

        self.assertTrue(self.Pipeline.is_error())
        self.assertEqual([event for context, event in output], [])



import os
import sys
import time
import signal
import random
import logging
import asyncio
import argparse
import itertools
import platform
import datetime
import typing

import bspump.asab as asab

try:
    import daemon
    import daemon.pidfile
    import lockfile
except ImportError:
    daemon = None

from .config import Config
from .abc.singleton import Singleton
from .log import Logging, _loop_exception_handler, LOG_NOTICE
from .task import TaskService

L = logging.getLogger(__name__)


class Application(metaclass=Singleton):
    """The base application object that maintains the global application state.

    You can provide your own implementation by creating a subclass. It is intended to be a Singleton.

    Examples:
    ```python
    class MyApplication(asab.Application):
            async def main(self):
                    print("Hello world!")
                    self.stop()

    if __name__ == "__main__":
            app = MyApplication()
            app.run()
    ```
    """

    Description = "This app is based on ASAB."

    def __init__(self, args: typing.Optional[list] = None, modules: list = []):
        """
        Initialize the Application provided with arguments and modules.

        Args:
                args: sequence of arguments to be parsed by `Application.parse_arguments()` call.
                modules: list of ASAB modules to be added by `Application.add_module()` call.

        Examples:

        ```python
        class MyApplication(asab.Application):
                def __init__(self):
                        super().__init__(modules=[asab.web.Module, asab.zookeeper.Module])
        ```
        """

        self.ExitCode: typing.Union[int, str]
        """
		The actual value of the exit code that can be set via `set_exit_code()` method.

		Examples:
			The example of the exit code handling in the `main()` function of the application:

			```python
			if __name__ == '__main__':
				app = asab.Application()
				exit_code = app.run()
				sys.exit(exit_code)
			```

			| Exit code | Meaning |
			| --- | --- |
			| 0 | success |
			| 1 | abnormal termination of a program perhaps as a result a minor problem in the code |
			| "!RESTART!" | hard restart of the whole application |

		"""

        try:
            # EX_OK code is not available on Windows
            self.ExitCode = os.EX_OK
        except AttributeError:
            self.ExitCode = 0

        # Queue of Services to be initialized
        self.InitServicesQueue = []
        # Queue of Modules to be initialized
        self.InitModulesQueue = []

        # Parse command line
        self.Args = self.parse_arguments(args=args)

        # Load configuration

        # Obtain HostName
        if deployment_id := os.environ.get("DEPLOYMENT_ID"):
            self.HostName = deployment_id
        elif container_id_file := os.environ.get("CONTAINER_ID_FILE"):
            while True:
                try:
                    with open(container_id_file, "r") as f:
                        self.HostName = f.read().strip()
                        break
                except FileNotFoundError:
                    time.sleep(1)
        else:
            self.HostName = platform.node()
        # Obtain DeploymentId
        self.DeploymentId = os.environ.get("DEPLOYMENT_ID", None)

        os.environ["HOSTNAME"] = self.HostName
        Config._load()

        if hasattr(self.Args, "daemonize") and self.Args.daemonize:
            self.daemonize()

        elif hasattr(self.Args, "kill") and self.Args.kill:
            self.daemon_kill()

        # Seed the random generator
        random.seed()

        # Obtain the event loop
        self.Loop = asyncio.get_event_loop()
        if self.Loop.is_closed():
            self.Loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.Loop)

        self.LaunchTime = time.time()
        self.BaseTime = self.LaunchTime - self.Loop.time()

        self.Modules: list[asab.Module] = []
        """
		A list of modules that has been added to the application.
		"""

        self.Services: dict[str, asab.Service] = {}
        """
		A dictionary of registered services.
		"""

        # Setup logging
        self.Logging = Logging(self)

        # Configure the event loop
        self.Loop.set_exception_handler(_loop_exception_handler)
        if Config["logging"].getboolean("verbose"):
            self.Loop.set_debug(True)

        # Adding a handler to listen to the interrupt event
        if platform.system() == "Windows":
            try:
                # Windows win32api import
                import win32api

                def handler(type):
                    self.stop()
                    return True

                win32api.SetConsoleCtrlHandler(handler, True)

            except ImportError as e:
                L.warning("win32api module could not be loaded, because '{}'".format(e))

        else:
            # POSIX and other reasonable systems
            self.Loop.add_signal_handler(signal.SIGINT, self.stop)
            self.Loop.add_signal_handler(signal.SIGTERM, self.stop)
            self.Loop.add_signal_handler(signal.SIGHUP, self._hup)

        self._stop_event = asyncio.Event()
        self._stop_event.clear()
        self._stop_counter = 0

        from .pubsub import PubSub

        self.PubSub = PubSub(self)

        L.info("Initializing ...")

        self.TaskService = TaskService(self)

        for module in modules:
            self.add_module(module)

        # Set housekeeping time and time limit
        (
            self.HousekeepingTime,
            self.HousekeepingTimeLimit,
            self.HousekeepingId,
        ) = self._initialize_housekeeping_schedule()
        self.HousekeepingMissedEvents: list = []
        # Every 10 minutes listen for housekeeping
        self.PubSub.subscribe("Application.tick/600!", self._on_housekeeping_tick)

    def create_argument_parser(
        self,
        prog=None,
        usage=None,
        description=None,
        epilog=None,
        prefix_chars="-",
        fromfile_prefix_chars=None,
        argument_default=None,
        conflict_handler="error",
        add_help=True,
    ) -> argparse.ArgumentParser:
        """
        Create an `argparse.ArgumentParser` object supplemented by default ASAB arguments.

        This method can be overridden to adjust argparse configuration.
        Refer to the Python standard library to [`argparse.ArgumentParser`](https://docs.python.org/3/library/argparse.html?highlight=argumentparser#argumentparser-objects) for the details.
        """

        parser = argparse.ArgumentParser(
            prog=prog,
            usage=usage,
            description=description if description is not None else self.Description,
            epilog=epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            prefix_chars=prefix_chars,
            fromfile_prefix_chars=fromfile_prefix_chars,
            argument_default=argument_default,
            conflict_handler=conflict_handler,
            add_help=add_help,
        )
        parser.add_argument(
            "-c", "--config", help="specify a path to a configuration file"
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="print more information (enable debug output)",
        )
        parser.add_argument(
            "-s", "--syslog", action="store_true", help="enable logging to a syslog"
        )
        parser.add_argument("-l", "--log-file", help="specify a path to a log file")
        parser.add_argument(
            "-w",
            "--web-api",
            help="activate Asab web API (default listening port is 0.0.0.0:8080)",
            const="0.0.0.0:8080",
            nargs="?",
        )
        parser.add_argument(
            "--startup-housekeeping",
            help="trigger housekeeping event immediately after application startup",
        )
        parser.add_argument(
            "--no-auth",
            action="store_true",
            help="disable all authentication and authorization",
        )

        if daemon is not None:
            parser.add_argument(
                "-d",
                "--daemonize",
                action="store_true",
                help="run daemonized (in the background)",
            )
            parser.add_argument(
                "-k",
                "--kill",
                action="store_true",
                help="kill a running daemon and quit",
            )

        return parser

    def parse_arguments(self, args=None):
        """
        Parse the command line arguments and set the default values for the configuration accordingly.

        Args:
                args: The arguments to parse. If not set, sys.argv[1:] will be used.

        Returns:
                The arguments that were parsed.
        """

        parser = self.create_argument_parser()
        args = parser.parse_args(args=args)

        if args.config is not None:
            Config._default_values["general"]["config_file"] = args.config

        if args.verbose:
            Config._default_values["logging"]["verbose"] = True

        if args.syslog:
            Config._default_values["logging:syslog"]["enabled"] = True

        if args.log_file:
            Config._default_values["logging:file"]["path"] = args.log_file

        if args.web_api:
            if "web" not in Config._default_values:
                Config._default_values["web"] = {}
            Config._default_values["web"]["listen"] = args.web_api

        if args.no_auth:
            if "auth" not in Config._default_values:
                Config._default_values["auth"] = {}
            Config._default_values["auth"]["enabled"] = False

        if args.startup_housekeeping:
            Config._default_values["housekeeping"]["run_at_startup"] = True

        return args

    def get_pidfile_path(self) -> typing.Optional[str]:
        """
        Get the path for PID file from the configuration.

        PID file is a file that contains process id of the ASAB process.
        It is used for interaction with OS respective it's control of running services.

        - If the `pidfile` is set to the empty string, return None.
        - If `pidfile` is set to "!", return the default PID file path (in `/var/run/` folder).
        This is the default value.

        Example of PID path configuration:

        ```ini
        [general]
        pidfile=/tmp/my.pid
        ```

        Returns:
                The path to the `pidfile`.
        """

        pidfilepath = Config["general"]["pidfile"]
        if pidfilepath == "":
            return None
        elif pidfilepath == "!":
            return os.path.join("/var/run", os.path.basename(sys.argv[0]) + ".pid")
        else:
            return pidfilepath

    def daemonize(self):
        if daemon is None:
            print(
                "Install 'python-daemon' module to support daemonizing.",
                file=sys.stderr,
            )
            sys.exit(1)

        pidfilepath = self.get_pidfile_path()
        if pidfilepath is not None:
            pidfile = daemon.pidfile.TimeoutPIDLockFile(pidfilepath)

        working_dir = Config["general"]["working_dir"]

        uid = Config["general"]["uid"]
        if uid == "":
            uid = None

        gid = Config["general"]["gid"]
        if gid == "":
            gid = None

        signal_map = {
            signal.SIGTTIN: None,
            signal.SIGTTOU: None,
            signal.SIGTSTP: None,
        }

        self.DaemonContext = daemon.DaemonContext(
            working_directory=working_dir,
            signal_map=signal_map,
            pidfile=pidfile,
            uid=uid,
            gid=gid,
        )

        try:
            self.DaemonContext.open()
        except lockfile.AlreadyLocked as e:
            print(
                "Cannot create a PID file '{}':".format(pidfilepath), e, file=sys.stderr
            )
            sys.exit(1)

    def daemon_kill(self):
        if daemon is None:
            print(
                "Install 'python-daemon' module to support daemonising.",
                file=sys.stderr,
            )
            sys.exit(1)

        pidfilepath = self.get_pidfile_path()
        if pidfilepath is None:
            sys.exit(0)

        try:
            pid = open(pidfilepath, "r").read()
        except FileNotFoundError:
            print("Pid file '{}' not found.".format(pidfilepath), file=sys.stderr)
            sys.exit(0)

        pid = int(pid)

        for sno in [
            signal.SIGINT,
            signal.SIGINT,
            signal.SIGINT,
            signal.SIGINT,
            signal.SIGTERM,
        ]:
            try:
                os.kill(pid, sno)
            except ProcessLookupError:
                print("Process with pid '{}' not found.".format(pid), file=sys.stderr)
                sys.exit(0)
            for i in range(10):
                if not os.path.exists(pidfilepath):
                    sys.exit(0)
                time.sleep(0.1)
            print(
                "Daemon process (pid: {}) still running ...".format(pid),
                file=sys.stderr,
            )

        print("Pid file '{}' not found.".format(pidfilepath), file=sys.stderr)
        sys.exit(1)

    def run(self):
        """Run the application.

        Returns:
                (int): Exit code of the finalized process.
        """

        # Commence init-time
        self.PubSub.publish("Application.init!")
        self.Loop.run_until_complete(
            asyncio.gather(
                self._init_time_governor(),
                self.initialize(),
            )
        )

        try:
            # Commence run-time and application main() function
            L.log(LOG_NOTICE, "is ready.")
            self._stop_event.clear()
            self.Loop.run_until_complete(
                asyncio.gather(
                    self._run_time_governor(),
                    self.main(),
                )
            )

            # Comence exit-time
            if self.ExitCode == "!RESTART!":
                L.log(LOG_NOTICE, "is restarting ...")
            else:
                L.log(
                    LOG_NOTICE,
                    "is exiting ...",
                    struct_data={"exit_code": self.ExitCode},
                )
            self.Loop.run_until_complete(
                asyncio.gather(
                    self.finalize(),
                    self._exit_time_governor(),
                )
            )

            # Python 3.5 lacks support for shutdown_asyncgens()
            if hasattr(self.Loop, "shutdown_asyncgens"):
                self.Loop.run_until_complete(self.Loop.shutdown_asyncgens())
            self.Loop.close()

        finally:
            if self.ExitCode == "!RESTART!":
                os.execv(sys.executable, [os.path.basename(sys.executable)] + sys.argv)

        return self.ExitCode

    def stop(self, exit_code: typing.Optional[int] = None) -> None:
        """
        Gracefully terminate the _run-time_ and commence the _exit-time_.

        This method is automatically called by `SIGINT` and `SIGTERM`.
        It also includes a response to `Ctrl-C` on UNIX-like system.
        When this method is called 3x, it abruptly exits the application (aka emergency abort).

        Args:
                exit_code (int, optional): Exit code of the finalized process.
        """
        if exit_code is not None:
            self.set_exit_code(exit_code)

        self._stop_event.set()
        self._stop_counter += 1
        self.PubSub.publish("Application.stop!", self._stop_counter)

        if self._stop_counter >= 3:
            L.fatal("Emergency exit")
            for task in asyncio.all_tasks():
                L.warning("Pending task during emergency exit: {}".format(task))
            try:
                # EX_SOFTWARE code is not available on Windows
                return os._exit(os.EX_SOFTWARE)
            except AttributeError:
                return os._exit(0)

        elif self._stop_counter > 1:
            L.warning("{} tasks still active".format(len(asyncio.all_tasks())))

    def _do_restart(self, event_name):
        self.stop("!RESTART!")

    def restart(self):
        """
        Schedule a hard restart of the whole application.

        This function works by using `os.execv()`, which replaces the current process with a new one (without creating a new process ID).
        Arguments and environment variables will be retained.

        !!! warning
                Please note that this will work on Unix-based systems only, as it uses a feature specific to Unix.

        !!! hint
                Be careful while using this function, make sure you have some control
                over when and how this function is being called to avoid any unexpected process restarts.
                It is not common to use these types of function calls in Python applications.
        """
        self.PubSub.subscribe("Application.tick/10!", self._do_restart)

    def _hup(self):
        self.Logging.rotate()
        self.PubSub.publish("Application.hup!")

    # Modules

    def add_module(self, module_class: asab.Module) -> None:
        """
        Load a new module.
        """

        for module in self.Modules:
            if isinstance(module, module_class):
                # Already loaded and registered
                return

        module = module_class(self)
        self.Modules.append(module)

        # Enqueue module for initialization (happens in run phase)
        self.InitModulesQueue.append(module)

    # Services

    def get_service(self, service_name: str) -> typing.Optional[asab.Service]:
        """
        Get a new service by its name.

        Args:
                service_name: Name of the service to retrieve.

        Returns:
                The service object associated with the provided service_name,
                or None if the service is not registered.
        """
        return self.Services.get(service_name)

    def _register_service(self, service: asab.Service):
        """
        Register a new service using its name.
        """

        if service.Name in self.Services:
            L.error(
                "Service '{}' already registered (existing:{} new:{})".format(
                    service.Name, self.Services[service.Name], service
                )
            )
            raise RuntimeError("Service {} already registered".format(service.Name))

        self.Services[service.Name] = service

        # Enqueue service for initialization (happens in run phase)
        self.InitServicesQueue.append(service)

    # Lifecycle callback

    async def initialize(self):
        """
        This method is called during the application *init-time*. It is intended to be overridden by the user.
        """
        pass

    async def main(self):
        """
        This method is called during the application *run-time*. It is intended to be overridden by the user.
        """
        pass

    async def finalize(self):
        """
        This method is called during the application *exit-time*. It is intended to be overridden by the user.
        """
        pass

    # Governors

    async def _init_time_governor(self):
        """
        Initialize all services that has been created during application construction
        """
        await self._ensure_initialization()

    async def _run_time_governor(self):
        timeout = Config.getint("general", "tick_period")
        self.PubSub.publish("Application.run!")

        if Config.getboolean("housekeeping", "run_at_startup", fallback=False):
            L.log(asab.LOG_NOTICE, "Startup housekeeping...")
            self.PubSub.publish("Application.housekeeping!")

        # Wait for stop event & tick in meanwhile
        for cycle_no in itertools.count(1):
            await self._ensure_initialization()

            try:
                await asyncio.wait_for(self._stop_event.wait(), timeout=timeout)
                break
            except asyncio.TimeoutError:
                self.PubSub.publish("Application.tick!")
                if (cycle_no % 10) == 0:
                    self.PubSub.publish("Application.tick/10!")
                if (cycle_no % 60) == 0:
                    # Rebase a Loop time
                    self.BaseTime = time.time() - self.Loop.time()
                    self.PubSub.publish("Application.tick/60!")
                if (cycle_no % 300) == 0:
                    self.PubSub.publish("Application.tick/300!")
                if (cycle_no % 600) == 0:
                    self.PubSub.publish("Application.tick/600!")
                if (cycle_no % 1800) == 0:
                    self.PubSub.publish("Application.tick/1800!")
                if (cycle_no % 3600) == 0:
                    self.PubSub.publish("Application.tick/3600!")
                if (cycle_no % 43200) == 0:
                    self.PubSub.publish("Application.tick/43200!")
                if (cycle_no % 86400) == 0:
                    self.PubSub.publish("Application.tick/86400!")
                continue

    async def _exit_time_governor(self):
        self.PubSub.publish("Application.exit!")

        # Finalize services
        futures = set()
        for service in self.Services.values():
            futures.add(asyncio.ensure_future(service.finalize(self)))

        while len(futures) > 0:
            done, futures = await asyncio.wait(
                futures, return_when=asyncio.FIRST_EXCEPTION
            )
            for fut in done:
                fut.result()

        # Finalize modules
        futures = set()
        for module in self.Modules:
            futures.add(asyncio.ensure_future(module.finalize(self)))

        while len(futures) > 0:
            done, futures = await asyncio.wait(
                futures, return_when=asyncio.FIRST_EXCEPTION
            )
            for fut in done:
                fut.result()

        # Wait for non-finalized tasks
        tasks_awaiting = 0
        for i in range(3):
            try:
                ts = asyncio.all_tasks(self.Loop)
            except AttributeError:
                # Compatibility for Python 3.6-
                ts = asyncio.Task.all_tasks(self.Loop)
            tasks_awaiting = 0
            for t in ts:
                if t.done():
                    continue
                tasks_awaiting += 1
            if tasks_awaiting <= 1:
                # 2 is for _exit_time_governor and wait()
                break

            await asyncio.sleep(1)

        else:
            L.warning(
                "Exiting but {} async task(s) are still waiting".format(tasks_awaiting)
            )

    async def _ensure_initialization(self):
        """
        This method ensures that any newly add module or registered service is initialized.
        It is called from:
        (1) init-time for modules&services added during application construction.
        (2) run-time for modules&services added during aplication lifecycle.
        """

        # Initialize modules
        while len(self.InitModulesQueue) > 0:
            module = self.InitModulesQueue.pop()
            await module.initialize(self)

        # Initialize services
        while len(self.InitServicesQueue) > 0:
            service = self.InitServicesQueue.pop()
            await service.initialize(self)

    def set_exit_code(self, exit_code: typing.Union[int, str], force: bool = False):
        """
        Set the exit code of the application.

        If `force` is `False`, the exit code will be set only if the previous value is lower than the new one.
        If `force` is `True`, the exit code value is set to `exit_code` value disregarding the previous value.

        Args:
                exit_code (str | int): The exit code value.
                force: Force the exit code reassignment.
        """
        if self.ExitCode == "!RESTART!":
            return

        if exit_code == "!RESTART!":
            self.ExitCode = exit_code

        elif (self.ExitCode < exit_code) or force:
            L.debug("Exit code set to {}".format(exit_code))
            self.ExitCode = exit_code

    # Time

    def time(self) -> float:
        """
        Return UTC UNIX timestamp using a loop time (a fast way how to get a wall clock time).

        Returns:
                Current UTC UNIX timestamp.
        """
        return self.BaseTime + self.Loop.time()

    # Housekeeping

    def _initialize_housekeeping_schedule(self):
        """
        Set the next housekeeping time and time limit from configuration.

        Returns:
                (next_housekeeping_time, next_time_limit, next_housekeeping_id)
        """
        config_house_time = datetime.datetime.strptime(
            Config["housekeeping"]["at"], "%H:%M"
        )  # default: 03:00
        config_time_limit = datetime.datetime.strptime(
            Config["housekeeping"]["limit"], "%H:%M"
        )  # default: 05:00

        now = datetime.datetime.now(datetime.timezone.utc)

        next_housekeeping_time = now.replace(
            hour=config_house_time.hour,
            minute=config_house_time.minute,
            second=0,
            microsecond=0,
        )

        # if the app started after the housekeeping time, set it to the next day
        if now > next_housekeeping_time:
            next_housekeeping_time += datetime.timedelta(days=1)

        # compute the time limit for the housekeeping
        time_delta_limit = config_time_limit - config_house_time
        if time_delta_limit < datetime.timedelta(hours=0):
            time_delta_limit += datetime.timedelta(days=1)

        next_time_limit = next_housekeeping_time + time_delta_limit

        # Each time has its id that prevents from accidental executing housekeeping twice.
        next_housekeeping_id = _housekeeping_id(now)

        return (next_housekeeping_time, next_time_limit, next_housekeeping_id)

    def _on_housekeeping_tick(self, message_type):
        """
        Check if it's time for publishing the 'Application.housekeeping!' message.
        If so, publish the message and set housekeeping time, the time limit and time id for the next day.
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        today_id = _housekeeping_id(now)

        if self.HousekeepingTime < now:
            if now < self.HousekeepingTimeLimit and self.HousekeepingId <= today_id:
                self.PubSub.publish("Application.housekeeping!")
            else:
                L.warning(
                    "Housekeeping has not been executed: It is past the time limit.",
                    struct_data={
                        "housekeeping_time": self.HousekeepingTime.strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        "time_limit": self.HousekeepingTimeLimit.strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        "housekeeping_id": self.HousekeepingId,
                    },
                )
                self.HousekeepingMissedEvents.append(today_id)

            self.HousekeepingTime += datetime.timedelta(days=1)
            self.HousekeepingTimeLimit += datetime.timedelta(days=1)
            self.HousekeepingId = _housekeeping_id(self.HousekeepingTime)
            L.log(
                LOG_NOTICE,
                "Setting time for the next housekeeping.",
                struct_data={
                    "next_housekeeping_time": self.HousekeepingTime.strftime(
                        "%Y-%m-%d %H:%M:%S"
                    ),
                    "next_time_limit": self.HousekeepingTimeLimit.strftime(
                        "%Y-%m-%d %H:%M:%S"
                    ),
                    "next_housekeeping_id": self.HousekeepingId,
                },
            )

            if len(self.HousekeepingMissedEvents) > 0:
                L.warning(
                    "One or more Housekeeping events have not been executed.",
                    struct_data={
                        "missed_housekeeping_events": self.HousekeepingMissedEvents
                    },
                )


def _housekeeping_id(dt: datetime.datetime) -> int:
    """
    Create a unique ID for each date. Utility function for housekeeping.

    >>> housekeeping_id(datetime.datetime.now())
    20230418
    """
    return int(dt.strftime("%Y%m%d"))



import codecs


class SourceProtocolABC(object):
    """
    Source protocol is a handler class, that basically gets the socket (in reader)
    and extract the payload from it in a way that is conformant to expected protocol.

    That is happening in the `handle()` method.
    The output is to be shipped to source.process() method.
    """

    def __init__(self, app, pipeline, config):
        """
        Description:

        """
        self.Loop = app.Loop

    async def handle(self, source, stream, context):
        """
        Description:


        """
        raise NotImplementedError()


class LineSourceProtocol(SourceProtocolABC):
    """
    Description: Basically readline() for reading lines from a socket.
    """

    def __init__(self, app, pipeline, config):
        """
        Description:


        """
        super().__init__(app, pipeline, config)

        # TODO: All following values could be read from configuration
        self.EOL = b"\n"
        self.SaneBufferSize = 64 * 1024  # The maximum buffer size considered as sane

        # Line decoder
        decode_codec = config["decode"]
        if decode_codec == "bytes":
            self.Codec = None
            self.LineDecoder = self._line_bytes_decoder
        else:
            self.Codec = codecs.lookup(decode_codec)
            self.LineDecoder = self._line_codec_decoder

    async def handle(self, source, stream, context):
        """
        Description:


        """
        pipeline = source.Pipeline

        input_buffer = bytearray(b" " * 8)
        input_buffer_mv = memoryview(input_buffer)
        input_buffer_pos = 0
        last_eol_pos = 0

        while True:
            recv_bytes = await stream.recv_into(input_buffer_mv[input_buffer_pos:])
            if recv_bytes <= 0:
                # Client closed the connection
                if recv_bytes < 0:
                    raise RuntimeError(
                        "Client sock_recv_into returned {}".format(recv_bytes)
                    )
                return

            input_buffer_pos += recv_bytes
            if len(input_buffer) == input_buffer_pos:
                if len(input_buffer) > self.SaneBufferSize:
                    raise RuntimeError("Insane buffer size requested")
                # Grow the input_buffer if the size touches the top
                new_input_buffer = bytearray(b" " * (len(input_buffer) * 2))
                input_buffer_mv = memoryview(new_input_buffer)
                input_buffer_mv[
                    : len(input_buffer)
                ] = input_buffer  # Copy the content of the old buffer
                input_buffer = new_input_buffer

            while last_eol_pos < input_buffer_pos:
                # Seek for end of line symbol in the buffer
                eol_pos = input_buffer[:input_buffer_pos].find(self.EOL, last_eol_pos)
                if eol_pos == -1:
                    break

                line = self.LineDecoder(line_bytes=input_buffer[last_eol_pos:eol_pos])
                last_eol_pos = eol_pos + 1

                await pipeline.ready()
                await source.process(line, context=context.copy())

            # TODO: HIGH PRIORITY This definitively doesn't cover all possible cases

            # If the '\n' is at the end of the buffer, reset the buffer position
            if last_eol_pos == input_buffer_pos:
                input_buffer_pos = 0
                last_eol_pos = 0
                continue

    def _line_codec_decoder(self, line_bytes):
        line, _ = self.Codec.decode(line_bytes)
        return line

    def _line_bytes_decoder(self, line_bytes):
        return line_bytes



import os

import bspump.declarative
import bspump.unittest


class TestDeclarativeAnd(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_dict_01(self):
        decl = self.load("./test_and_01.yaml")

        bspump.declarative.declaration_to_dot(decl, "./test_and_01.dot")



import json


def load_json_file(fname):
    if fname.endswith(".gz"):
        import gzip

        f = gzip.open(fname, "rb")

    elif fname.endswith(".bz2"):
        import bz2

        f = bz2.open(fname, "rb")

    elif fname.endswith(".xz") or fname.endswith(".lzma"):
        import lzma

        f = lzma.open(fname, "rb")

    else:
        f = open(fname, "rb")

    jdata = json.loads(f.read().decode("utf-8"))
    f.close()
    return jdata



import abc
import secrets
import hashlib
import logging
from bspump.asab import Config, Service
import re

try:
    import cryptography.hazmat.primitives.ciphers
    import cryptography.hazmat.primitives.ciphers.algorithms
    import cryptography.hazmat.primitives.ciphers.modes
except ModuleNotFoundError:
    cryptography = None

#

L = logging.getLogger(__name__)

#


ENCRYPTED_PREFIX = b"$aes-cbc$"


class StorageServiceABC(Service):
    """
    An abstract class for the Storage Service.

    """

    def __init__(self, app, service_name):
        super().__init__(app, service_name)
        self.WebhookURIs = (
            Config.get("asab:storage:changestream", "webhook_uri", fallback="") or None
        )
        if self.WebhookURIs is not None:
            self.WebhookURIs = [
                uri for uri in re.split(r"\s+", self.WebhookURIs) if len(uri) > 0
            ]
            try:
                self.ProactorService = app.get_service("asab.ProactorService")
            except KeyError as e:
                raise Exception("Storage webhooks require ProactorService") from e
        self.WebhookAuth = (
            Config.get("asab:storage:changestream", "webhook_auth", fallback="") or None
        )

        # Specify a non-empty AES key to enable AES encryption of selected fields
        self._AESKey = Config.get("asab:storage", "aes_key", fallback="")
        if len(self._AESKey) > 0:
            if cryptography is None:
                raise ModuleNotFoundError(
                    "You are using storage encryption without 'cryptography' installed. "
                    "Please run 'pip install cryptography' "
                    "or install asab with 'storage_encryption' optional dependency."
                )
            self._AESKey = hashlib.sha256(self._AESKey.encode("utf-8")).digest()
        else:
            self._AESKey = None

    @abc.abstractmethod
    def upsertor(self, collection: str, obj_id=None, version: int = 0) -> None:
        """
		Create an upsertor object for the specified collection.

		If updating an existing object, please specify its `obj_id` and also `version` that you need to read from a storage upfront.
		If `obj_id` is None, we assume that you want to insert a new object and generate its new `obj_id`, `version` should be set to 0 (default) in that case.
		If you want to insert a new object with a specific `obj_id`, specify `obj_id` and set a version to 0.
			- If there will be a colliding object already stored in a storage, `execute()` method will fail on `DuplicateError`.

		Args:
			collection: Name of collection to work with
			obj_id: Primary identification of an object in the storage (e.g. primary key)
			version: Specify a current version of the object and hence prevent byzantine faults. \
			You should always read the version from the storage upfront, prior using an upsertor. \
			That creates a soft lock on the record. It means that if the object is updated by other \
			component in meanwhile, your upsertor will fail and you should retry the whole operation. \
			The new objects should have a `version` set to 0.
		"""
        pass

    @abc.abstractmethod
    async def get(self, collection: str, obj_id, decrypt: bool = None) -> dict:
        """
        Get object from collection by its ID.

        Args:
                collection: Collection to get from.
                obj_id: Object identification.
                decrypt: Set of fields to decrypt.

        Returns:
                The object retrieved from a storage.

        Raises:
                KeyError: Raised if `obj_id` is not found in `collection`.
        """
        pass

    @abc.abstractmethod
    async def get_by(self, collection: str, key: str, value, decrypt=None) -> dict:
        """
        Get object from collection by its key and value.

        Args:
                collection: Collection to get from
                key: Key to filter on
                value: Value to filter on
                decrypt: Set of fields to decrypt

        Returns:
                The object retrieved from a storage.

        Raises:
                KeyError: If object {key: value} not found in `collection`
        """
        pass

    @abc.abstractmethod
    async def delete(self, collection: str, obj_id):
        """
        Delete object from collection.

        Args:
                collection: Collection to get from
                obj_id: Object identification

        Returns:
                ID of the deleted object.

        Raises:
                KeyError: Raised when obj_id cannot be found in collection.
        """
        pass

    def aes_encrypt(self, raw: bytes, iv: bytes = None) -> bytes:
        """
        Take an array of bytes and encrypt it using AES-CBC.

        Args:
                raw: The data to be encrypted.
                iv: AES-CBC initialization vector, 16 bytes long. If left empty, a random 16-byte array will be used.

        Returns:
                The encrypted data.

        Raises:
                TypeError: The data are not in binary format.
        """
        block_size = (
            cryptography.hazmat.primitives.ciphers.algorithms.AES.block_size // 8
        )

        if self._AESKey is None:
            raise RuntimeError(
                "No aes_key specified in asab:storage configuration. "
                "If you want to use encryption, specify a non-empty aes_key."
            )

        if not isinstance(raw, bytes):
            if isinstance(raw, str):
                raise TypeError("String objects must be encoded before encryption")
            else:
                raise TypeError("Only 'bytes' objects can be encrypted")

        # Pad the text to fit the blocks
        pad_length = -len(raw) % block_size
        if pad_length != 0:
            raw = raw + b"\00" * pad_length

        if iv is None:
            iv = secrets.token_bytes(block_size)

        algorithm = cryptography.hazmat.primitives.ciphers.algorithms.AES(self._AESKey)
        mode = cryptography.hazmat.primitives.ciphers.modes.CBC(iv)
        cipher = cryptography.hazmat.primitives.ciphers.Cipher(algorithm, mode)
        encryptor = cipher.encryptor()
        encrypted = (
            ENCRYPTED_PREFIX + iv + (encryptor.update(raw) + encryptor.finalize())
        )
        return encrypted

    def aes_decrypt(self, encrypted: bytes) -> bytes:
        """
        Decrypt encrypted data using AES-CBC.

        Args:
                encrypted: The encrypted data to decrypt. It must start with b"$aes-cbc$" prefix, followed by one-block-long initialization vector.

        Returns:
                The decrypted data.
        """
        block_size = (
            cryptography.hazmat.primitives.ciphers.algorithms.AES.block_size // 8
        )

        if self._AESKey is None:
            raise RuntimeError(
                "No aes_key specified in asab:storage configuration. "
                "If you want to use encryption, specify a non-empty aes_key."
            )

        if not isinstance(encrypted, bytes):
            raise TypeError("Only values of type 'bytes' can be decrypted")

        # Strip the prefix
        if not encrypted.startswith(ENCRYPTED_PREFIX):
            raise ValueError(
                "Encrypted data must start with {!r} prefix".format(ENCRYPTED_PREFIX)
            )
        encrypted = encrypted[len(ENCRYPTED_PREFIX) :]

        # Separate the initialization vector
        iv, encrypted = encrypted[:block_size], encrypted[block_size:]

        algorithm = cryptography.hazmat.primitives.ciphers.algorithms.AES(self._AESKey)
        mode = cryptography.hazmat.primitives.ciphers.modes.CBC(iv)
        cipher = cryptography.hazmat.primitives.ciphers.Cipher(algorithm, mode)
        decryptor = cipher.decryptor()
        raw = decryptor.update(encrypted) + decryptor.finalize()

        # Strip padding
        raw = raw.rstrip(b"\x00")
        return raw

    def encryption_enabled(self) -> bool:
        """
        Check if AESKey is not empty.

        Returns:
                True if AESKey is not empty.
        """
        return self._AESKey is not None



from bspump.main import NotebookCompiler
import json
import ast
import pytest


def test_notebook_parse_valid():
    compiler = NotebookCompiler()
    with open("jupyter/parse_example.ipynb", "r") as ntbf:
        ntb = json.load(ntbf)
    compiler.compile_notebook(ntb, "jupyter/tmp.py")
    with open("jupyter/tmp.py", "r") as outf:
        out = ast.parse(outf.read())
    with open("jupyter/expected.py", "r") as expectf:
        expect = ast.parse(expectf.read())
    assert ast.dump(out) == ast.dump(expect)


def test_notebook_formatting():
    compiler = NotebookCompiler()
    with open("jupyter/parse_example.ipynb", "r") as ntbf:
        ntb = json.load(ntbf)
    compiler.compile_notebook(ntb, "jupyter/tmp.py")
    with open("jupyter/tmp.py", "r") as outf:
        out = outf.read()
    with open("jupyter/expected.py", "r") as expectf:
        expect = expectf.read()
    assert out == expect



import re
import logging
import asyncio
import aiohttp
from ...abc.source import TriggerSource


L = logging.getLogger(__name__)


class InvalidResponseStatusCodeError(aiohttp.ClientError):
    pass


class HTTPABCClientSource(TriggerSource):
    ConfigDefaults = {
        "method": "GET",
        "url": "http://example.com/",
        "response_code": "200",  # Specify an expected response status code, more values are accepted ("200, 300 301")
        "max_failed_retries": 3,
        "fail_chilldown": 30,  # If there is an incorrect response, chilldown for X seconds
    }

    def __init__(self, app, pipeline, id=None, config=None, headers={}):
        super().__init__(app, pipeline, id=id, config=config)
        self.Loop = app.Loop

        self.Method = self.Config["method"]
        self.URL = self.Config["url"]

        self.Headers = headers.copy()

        self.SSL = None
        # SSL validation mode (see aiohttp documentation for more details)
        # - None for default SSL check (ssl.create_default_context() is used),
        # - False for skip SSL certificate validation
        # - aiohttp.Fingerprint for fingerprint validation
        # - ssl.SSLContext for custom SSL certificate validation.

        self.FailedResponses = 0
        self.MaxFailedResponses = int(self.Config.get("max_failed_retries"))
        self.FailChilldown = float(self.Config.get("fail_chilldown"))
        try:
            self.ResponseCodes = frozenset(
                map(int, re.findall(r"\d+", self.Config.get("response_code")))
            )
        except Exception:
            L.error("Failed to parse 'response_code' configuration value")
            raise

    async def main(self):
        async with aiohttp.ClientSession() as session:
            await super().main(session)

    async def cycle(self, session):
        try:
            async with session.request(
                self.Method,
                self.URL,
                headers=self.Headers if len(self.Headers) > 0 else None,
                ssl=self.SSL,
            ) as response:
                if response.status not in self.ResponseCodes:
                    await response.text()
                    raise InvalidResponseStatusCodeError(
                        "The response status code {} from '{}' is invalid".format(
                            response.status, self.URL
                        )
                    )
                else:
                    self.FailedResponses = 0  # Reset the counter
                await self.read(response)

        except aiohttp.ClientError as e:
            self.FailedResponses += 1
            if self.FailedResponses < self.MaxFailedResponses:
                L.warning(
                    "{}, will retry ({}/{}) in {:0.0f} sec".format(
                        e,
                        self.FailedResponses,
                        self.MaxFailedResponses,
                        self.FailChilldown,
                    )
                )

                self.Pipeline.MetricsEPSCounter.add("warning", 1)
                self.Pipeline.MetricsCounter.add("warning", 1)

                self.Pipeline.MetricsEPSCounter.add("eps.in", 1)
                self.Pipeline.MetricsCounter.add("event.in", 1)
                await asyncio.sleep(self.FailChilldown)
            else:
                L.error("{}, {} failed response(s)".format(e, self.FailedResponses))
                raise aiohttp.ClientError("{}, url='{}'".format(e, self.URL))

    async def read(self, response):
        """
        Override this method to implement your HTTP Source.
        """
        raise NotImplementedError()



import logging

from .analyzer import Analyzer
from ..matrix.timewindowmatrix import TimeWindowMatrix, PersistentTimeWindowMatrix

###

L = logging.getLogger(__name__)

###


class TimeWindowAnalyzer(Analyzer):
    """
    This is the analyzer for events with a temporal dimension (aka timestamp).
    Configurable sliding window records events withing specified windows and implements functions to find the exact time slot.
    Timer periodically shifts the window by time window resolution, dropping previous events.

    `TimeWindowAnalyzer` operates over the `TimeWindowMatrix` object.

    `resolution`is how many seconds fit in one time cell, default value is `60`.
    `start_time` is a unix timestamp for time to start. Default value is `None`, which will be equivalent current time.
    `clock_driven` is a boolean parameter, specifying how the matrix should be advanced. If `True`, it advances on timer's tick,
    else manually. Default value is `True`.
    `matrix_id` is an id of `TimeWindowMatrix` object alternatively passed, if not provided, the new matrix will be created with and ID derived from the Analyzer Id
    `analyze_on_clock` enables enables analyzis by timer.

    If the `TimeWindowAnalyzer` is `clock_driven`, the time should be periodically shifted (`on_clock_tick()`). The same
    function runs analyzis, if it's enabled.
    """

    def __init__(
        self,
        app,
        pipeline,
        matrix_id=None,
        dtype="float_",
        columns=15,
        analyze_on_clock=False,
        resolution=60,
        start_time=None,
        clock_driven=False,
        persistent=False,
        id=None,
        config=None,
    ):
        """
        Description:

        """

        super().__init__(
            app, pipeline, analyze_on_clock=analyze_on_clock, id=id, config=config
        )
        svc = app.get_service("bspump.PumpService")
        if matrix_id is None:
            matrix_id = self.Id + "Matrix"
            if persistent:
                self.TimeWindow = PersistentTimeWindowMatrix(
                    app,
                    dtype=dtype,
                    columns=columns,
                    resolution=resolution,
                    clock_driven=clock_driven,
                    start_time=start_time,
                    id=matrix_id,
                    config=config,
                )
            else:
                self.TimeWindow = TimeWindowMatrix(
                    app,
                    dtype=dtype,
                    columns=columns,
                    resolution=resolution,
                    clock_driven=clock_driven,
                    start_time=start_time,
                    id=matrix_id,
                    config=config,
                )
            svc.add_matrix(self.TimeWindow)
        else:
            # locate
            self.TimeWindow = svc.locate_matrix(matrix_id)



import logging
import os
import numpy as np

from bspump.asab import PubSub
from .utils.index import Index, PersistentIndex
from .matrix import Matrix, PersistentMatrix

###

L = logging.getLogger(__name__)

###


class NamedMatrix(Matrix):
    def __init__(self, app, dtype="float_", id=None, config=None):
        super().__init__(app, dtype=dtype, id=id, config=config)
        self.PubSub = PubSub(app)

    def zeros(self):
        super().zeros()
        self.Index = Index()

    def serialize(self):
        serialized = {}
        serialized["Index"] = self.Index.serialize()
        serialized["ClosedRows"] = self.ClosedRows.serialize()
        serialized["DType"] = self.DType
        serialized["Array"] = self.Array.tolist()
        return serialized

    def deserialize(self, data):
        self.Index.deserialize(data["Index"])
        self.ClosedRows.deserialize(data["ClosedRows"])

        if isinstance(data["DType"], str):
            self.DType = data["DType"]
        else:
            self.DType = []
            for member in data["DType"]:
                self.DType.append(tuple(member))

        array = []
        for member in data["Array"]:
            array.append(tuple(member))

        self.Array = np.array(array, dtype=self.DType)

    def _grow_rows(self, rows=1):
        super()._grow_rows(rows)
        self.Index.extend(self.Array.shape[0])

    def flush(self):
        """
        The matrix will be recreated without rows from `ClosedRows`.
        """
        closed_indexes, saved_indexes = super().flush()
        self.Index.flush(closed_indexes)
        return closed_indexes, saved_indexes

    def add_row(self, row_name: str):
        assert row_name is not None

        row_index = super().add_row()
        self.Index.add_row(row_name, row_index)
        self.PubSub.publish("Matrix changed!")
        return row_index

    def close_row(self, row_name, clear=True):
        row_index = self.Index.get_row_index(row_name)
        if row_index in self.ClosedRows:
            return False

        if row_index is None:
            return False

        self.Index.pop_index(row_index)
        self.PubSub.publish("Matrix changed!")

        if clear:
            self.Array[row_index] = np.zeros(1, dtype=self.DType)

        self.ClosedRows.add(row_index)
        if len(self.ClosedRows) >= self.MaxClosedRowsCapacity * self.Array.shape[0]:
            self.flush()

        crc = len(self.ClosedRows)
        self.Gauge.set("rows.active", self.Array.shape[0] - crc)
        self.Gauge.set("rows.closed", crc)
        return True

    def close_rows(self, row_names, clear=True):
        for name in row_names:
            self.close_row(name, clear=clear)

    def get_row_index(self, row_name: str):
        return self.Index.get_row_index(row_name)

    def get_row_name(self, row_index: int):
        return self.Index.get_row_name(row_index)


class PersistentNamedMatrix(PersistentMatrix):
    def __init__(self, app, dtype="float_", id=None, config=None):
        super().__init__(app, dtype=dtype, id=id, config=config)
        self.PubSub = PubSub(app)

    def zeros(self):
        super().zeros()
        path = os.path.join(self.Path, "map.dat")
        self.Index = PersistentIndex(path, self.Array.shape[0])

    def _grow_rows(self, rows=1):
        super()._grow_rows(rows)
        self.Index.extend(self.Array.shape[0])

    def flush(self):
        """
        The matrix will be recreated without rows from `ClosedRows`.
        """
        closed_indexes, saved_indexes = super().flush()
        self.Index.flush(closed_indexes)
        return closed_indexes, saved_indexes

    def add_row(self, row_name: str):
        assert row_name is not None

        row_index = super().add_row()
        self.Index.add_row(row_name, row_index)
        self.PubSub.publish("Matrix changed!")
        return row_index

    def close_row(self, row_name, clear=True):
        row_index = self.Index.get_row_index(row_name)
        if row_index in self.ClosedRows:
            return False

        if row_index is None:
            return False

        self.Index.pop_index(row_index)
        self.PubSub.publish("Matrix changed!")

        if clear:
            self.Array[row_index] = np.zeros(1, dtype=self.DType)

        self.ClosedRows.add(row_index)
        if len(self.ClosedRows) >= self.MaxClosedRowsCapacity * self.Array.shape[0]:
            self.flush()

        crc = len(self.ClosedRows)
        self.Gauge.set("rows.active", self.Array.shape[0] - crc)
        self.Gauge.set("rows.closed", crc)
        return True

    def close_rows(self, row_names, clear=True):
        for name in row_names:
            self.close_row(name, clear=clear)

    def get_row_index(self, row_name: str):
        return self.Index.get_row_index(row_name)

    def get_row_name(self, row_index: int):
        return self.Index.get_row_name(row_index)



from bspump.asab import Service, Application, Config
import sys
import os
import re
import logging
import json
import typing


L = logging.getLogger(__name__)


try:
    import sentry_sdk
    import sentry_sdk.integrations.aiohttp
    import sentry_sdk.integrations.asyncio
    import sentry_sdk.integrations.logging

    # IMPORTANT: This is because the default settings truncates long stack traces
    # https://stackoverflow.com/questions/53699110/how-do-i-increase-the-max-length-of-captured-python-parameters-in-sentry
    sentry_sdk.utils.MAX_STRING_LENGTH = 8192

except ModuleNotFoundError:
    L.critical(
        "Package for Sentry SDK not found. Install it with: pip install sentry-sdk"
    )
    sys.exit(1)


class SentryService(Service):
    """
    Service for Sentry SDK integration.

    Sentry is an error tracking and performance monitoring platform.
    When the service is initialized and `data_source_name` is set,
    all unhandled exceptions, error log messages are sent as Events to sentry.io,
    together with lines of code where the error happened, structured data and values of variables.

    Configuration:
    ```ini
    [asab:alert:sentry]
    data_source_name=... ; DSN of the project
    environment=... ; default: 'not specified'
    ```

    Examples:

    ```python
    class MyApp(asab.Application):
            async def initialize(self):
                    if "asab:alert:sentry" in asab.Config.sections():
                            import bspump.asab.sentry
                            self.SentryService = bspump.asab.sentry.SentryService(self)
    ```
    """

    def __init__(self, app: Application, service_name: str = "asab.SentryService"):
        super().__init__(app, service_name)

        # DATA SOURCE NAME (DSN)
        # format: https://<public key>@o<secret key>.ingest.sentry.io/<project id>
        # DSN is automatically generated when new project is created
        # and can be modified: Settings > Client Keys (DSN) > Key Details
        # Specification: either in configuration '[sentry] data_source_name', $SENTRY_DSN environment variable as a fallback
        self.DataSourceName = Config.get(
            "asab:alert:sentry", "data_source_name", fallback=""
        )
        if len(self.DataSourceName) == 0:
            self.DataSourceName = os.getenv("SENTRY_DSN", "")
        if len(self.DataSourceName) == 0:
            # We do not need Sentry enabled in development - empty DSN leads to failure
            L.critical(
                "Data source name is not set. Specify it in configuration: '[sentry] data_source_name'."
            )
            raise SystemExit("Exit due to a critical configuration error.")

        # ENVIRONMENT (e.g. "production", "testing", ...)
        self.Environment = Config.get(
            "asab:alert:sentry",
            "environment",
            fallback="development",
        )

        # RELEASE
        # Release can be obtained from MANIFEST.json if exists
        manifest = {}
        manifest_path = Config.get("general", "manifest", fallback="")
        if manifest_path == "":
            if os.path.isfile("/app/MANIFEST.json"):
                manifest_path = "/app/MANIFEST.json"
            elif os.path.isfile("/MANIFEST.json"):
                manifest_path = "/MANIFEST.json"
            elif os.path.isfile("MANIFEST.json"):
                manifest_path = "MANIFEST.json"

        if len(manifest_path) != 0:
            with open(manifest_path) as f:
                manifest = json.load(f)

        self.Release = (
            "{appname}:{version}".format(  # e.g. 'LMIOParsecApplication:v23.40-alpha'
                appname=app.__class__.__name__,
                version=manifest.get("version", "<none>"),
            )
        )

        # PERFORMANCE MONITORING
        # traces sample rate: percentage of captured events
        # prevents overcrowding when deployed to production
        # default: 100%
        self.TracesSampleRate = Config.getfloat(
            "asab:alert:sentry", "traces_sample_rate", fallback=1.0
        )
        assert (
            0 <= self.TracesSampleRate <= 1.0
        ), "Traces sample rate must be between 0 and 1."

        # INITIALIZATION
        sentry_sdk.init(
            dsn=self.DataSourceName,
            integrations=[
                sentry_sdk.integrations.aiohttp.AioHttpIntegration(),
                sentry_sdk.integrations.asyncio.AsyncioIntegration(),
            ],
            traces_sample_rate=self.TracesSampleRate,  # percentage of captured events
            environment=self.Environment,  # e.g. "production", "develop"
            release=self.Release,  # version of the microservice, e.g., v23.40-alpha
            auto_session_tracking=True,  # session info about interaction between user and app
            debug=False,  # ...sends many irrelevant messages
            max_value_length=8192,  # longer messages are truncated, the default value (1024) is too short
        )
        # TODO: Investigate CA certs, TLS/SSL, Security Tokens, Allowed Domains

        # ADDITIONAL GLOBAL TAGS
        # These tags will be set manually or automatically by Remote Control
        self.NodeId = os.getenv("NODE_ID", None)  # e.g. "lmio-box-testing-1"
        self.ServiceId = os.getenv("SERVICE_ID", None)  # e.g. "lmio-service"
        self.InstanceId = os.getenv("INSTANCE_ID", None)  # e.g. "lmio-service-01"

        if self.NodeId:
            sentry_sdk.set_tag("node_id", self.NodeId)
        if self.ServiceId:
            sentry_sdk.set_tag("service_id", self.ServiceId)
        if self.InstanceId:
            sentry_sdk.set_tag("instance_id", self.InstanceId)

        sentry_sdk.set_tag(
            "appclass", app.__class__.__name__
        )  # e.g. 'LMIOParsecApplication'

        L.info("is ready.")  # for debugging, visible only if argument '-v' is set

    def capture_exception(self, error=None, scope=None, **scope_args):
        """
        Capture caught exception and send it to Sentry.

        Args:
                error (str, optional): Error message that will be sent. If not specified, the one currently held in `sys.exc_info()` is sent.

        Examples:
        ```python
        try:
                call_collapsing_function()
        except Exception as e:
                sentry_svc.capture_exception(e)
        ```
        """
        return sentry_sdk.capture_exception(error=error, scope=scope, **scope_args)

    def capture_message(self, message, level=None, scope=None, **scope_args):
        """
        Send textual information to Sentry.

        Args:
                message (str):  Text message that will be sent.
        """
        return sentry_sdk.capture_message(
            message=message, level=level, scope=scope, **scope_args
        )

    def set_tag(self, key: str, value: typing.Union[str, int]) -> None:
        """
        Add custom tag to the current scope.

        Tags are key-value string pairs that are both indexed and searchable. They can help you quickly both access related events and view the tag distribution for a set of events.

        Tag is set only for the current scope (function, method, class, module).

        Args:
                key (str): Tag key. Tag keys can have a maximum length of 32 characters and can contain only letters (a-zA-Z), numbers (0-9), underscores (_), periods (.), colons (:), and dashes (-).
                value: Tag value. Tag values can have a maximum length of 200 characters and they cannot contain the newline (`\\n`) character.
        """

        # Check key format
        if not (0 < len(key) <= 32):
            L.error("Tag key '{}' is too long.".format(key))
            return

        key_pattern = re.compile("^[a-zA-Z0-9_.:-]+$")
        if key_pattern.match(key) is None:
            L.error("Tag '{}' contains invalid characters.".format(key))
            return

        # Check value format
        if not (0 < len(value) <= 200):
            L.error("Tag value '{}' is too long.".format(value))
            return

        if "\n" in value:
            L.error("Tag value {} contains '\\n' character.")
            return

        return sentry_sdk.set_tag(key, value)

    def set_tags(self, tags_dict: dict) -> None:
        """
        Add multiple custom tags to the current scope.

        Tags are key-value string pairs that are both indexed and searchable. They can help you quickly both access related events and view the tag distribution for a set of events.

        Tags are set only for the current scope (function, method, class, module).

        Args:
                tags_dict (dict): Dictionary of tag keys and values.
        """
        for key, value in tags_dict.items():
            self.set_tag(key, value)

    def transaction(self, span_operation: str, span_name: str):
        """
        Start a new transaction.

        Transactions are used for performance monitoring.

        This method is used as a context manager.

        Args:
                span_operation (str): Displayed span operation name that cannot be filtered, e.g., 'task'.
                span_name (str): Displayed span name that can be filtered.

        Returns:
                Transaction: A context manager that measures time operation of the task inside.

        Examples:
        ```python
        with sentry_svc.transaction("speed test", "test sleeping"):
                time.sleep(1.0)
        ```
        """
        return sentry_sdk.start_transaction(op=span_operation, name=span_name)

    def span(self, operation: str, description: str):
        """
        Create a child span within custom transaction.

        This method is used as a context manager.

        Args:
                operation (str): Displayed span operation name that cannot be filtered, e.g., 'task'.
                description (str): Displayed span name that can be filtered.

        Returns:
                Span: A context manager that measures time operation of the task inside.

        Examples:
        ```python
        with sentry_svc.transaction("speed test", "multiple tasks"):
                prepare_task1()
                with sentry_svc.span("task", "task1"):
                        task1()
                prepare_task2()
                with sentry_svc.span("task", "task2"):
                        task2()
                finalize()
        ```
        """
        return sentry_sdk.start_span(op=operation, description=description)

    def set_context(self, section_name: str, context: dict) -> None:
        """
        Set custom context within the current scope.

        Context data are related to the current user and the environment.
        They are useful when you have all the data available at a single point in time,
        not well suited to data that is collected over time.

        Args:
                section_name (str): A unique name for the context.
                context (dict): Key-value paired context dictionary.

        Examples:

        ```python
        sentry_service.set_context("environment variables", {"VAR": os.getenv("VAR")})
        ```

        !!! note
                We prefer using logging messages to setting custom context.
        """
        return sentry_sdk.set_context(section_name, context)



import asyncio
import json
import logging

import psycopg2
import psycopg2.extras

from ..abc.source import Source

#

L = logging.getLogger(__name__)

#


class PostgreSQLLogicalReplicationSource(Source):
    """
    This is the source, which reads Postgres WAL-file and
    produces events `INSERT`, `DELETE` or `UPDATE`.
    By default it uses `wal2json` postgresql plugin
    and it is not preinstalled. Here are the steps, how to install it.
    `git clone https://github.com/eulerto/wal2json`
    `cd wal2json`
    `make`
    `make install`
    Then update postgresql.conf with lines:
    ```
    ############ REPLICATION ##############
    # MODULES
    shared_preload_libraries = 'wal2json'

    # REPLICATION
    wal_level = logical
    max_wal_senders = 4
    max_replication_slots = 4
    ```
    Then restart postgres:
    `sudo service postgresql restart`

    You can use a different plug-in, but you have to override `decode()`.
    """

    ConfigDefaults = {
        "slot_name": "",
        "output_plugin": "wal2json",
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Loop = app.Loop
        self.ConnectionGeneral = pipeline.locate_connection(app, connection)
        self.DSN = self.ConnectionGeneral.build_dsn()
        self.SlotName = self.Config["slot_name"]
        self.OutputPlugin = self.Config["output_plugin"]
        self.Queue = asyncio.Queue(loop=self.Loop)
        self.ProactorService = app.get_service("asab.ProactorService")
        self.Running = True

    def decode(self, message):
        """
        Override it if you use a different plug-in
        or you want the output in non-dictionary form.
        """
        message = json.loads(message)
        return message

    def stream_data(self):
        while True:
            if not self.Running:
                return False

            mes = self.Cursor.read_message()
            if mes is not None:
                self.Queue.put_nowait(({}, self.decode(mes.payload)))

        return True

    async def main(self):
        await self.Pipeline.ready()

        conn = psycopg2.connect(
            self.DSN, connection_factory=psycopg2.extras.LogicalReplicationConnection
        )
        self.Cursor = conn.cursor()

        try:
            self.Cursor.start_replication(slot_name=self.SlotName, decode=False)
        except psycopg2.ProgrammingError:
            self.Cursor.create_replication_slot(
                self.SlotName, output_plugin=self.OutputPlugin
            )
            self.Cursor.start_replication(slot_name=self.SlotName, decode=False)

        self.ProactorService.execute(self.stream_data)

        try:
            while True:
                context, event = await self.Queue.get()
                await self.process(event, context={})
                self.Queue.task_done()

        except asyncio.CancelledError:
            if self.Queue.qsize() > 0:
                self.Running = False
                L.warning(
                    "'{}' stopped with {} events in a queue".format(
                        self.locate_address(), self.Queue.qsize()
                    )
                )

        finally:
            self.Running = False



import bspump.crypto
import bspump.unittest

# Classes


class SHA256HashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {"algorithm": "sha256"}


class SHA224HashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {"algorithm": "sha224"}


class SHA384HashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {"algorithm": "sha384"}


class SHA512HashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {"algorithm": "sha512"}


class SHA1HashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {"algorithm": "sha1"}


class MD5HashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {"algorithm": "md5"}


class BLAKE2BHashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {"algorithm": "blake2b"}


class BLAKE2SHashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {
        "algorithm": "blake2s",
        "digest_size": 32,
    }


class UnknownAlgorithmHashingProcessor(bspump.crypto.HashingProcessor):
    ConfigDefaults = {"algorithm": "__unknown__"}


class MD5ContextHashingProcessor(bspump.crypto.CoHashingProcessor):
    ConfigDefaults = {"algorithm": "md5"}


# Test


class TestSHA256HashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_sha256_hashing_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(SHA256HashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 32)
        self.assertEqual(
            output[0][1],
            b"\x86ex\x08\x1a\x1d\xeas\xbeG\xdc\xdeE3K\x17\xb2\xcc\xc0=\x88-\xa2\xd0\xfb\x1eQ\xec\x9d\xdcse",
        )


class TestSHA224HashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_sha224_hashing_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(SHA224HashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 28)
        self.assertEqual(
            output[0][1],
            b"\xfch\xfc\x9e*\x04\x83\x17\x9f\xd7'\xc63w\x9f\xb6\x01\xf2\n~\xa4?J\x84z\xcaCb",
        )


class TestSHA384HashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_sha384_hashing_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(SHA384HashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 48)
        self.assertEqual(
            output[0][1],
            b"\xf4\xee\x1d\x1e)\xaa_\x1dQ#\xcf\xf3\xf7\xfb \xeb\xe4\xca>z\x95\xf4\xddf_q\xd2\x9f\x93\xa7\x98n\xc6?\x92\x16\x8a2\x1b\x14\x7f\xc4e\ra<\xc9\x88",  # noqa
        )


class TestSHA512HashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_sha512_hashing_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(SHA512HashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 64)
        self.assertEqual(
            output[0][1],
            b"9~sN\x11\x0e \xe9-\xca\xd2O\xea\x8d\xdd\xc0t'^\xb8\xb8VE\xae\xe7\xe5\xa0<\xadc\xd2\x1c\xd3\t(\xb0\x84w\xe9U\x108a\xb3\xa1\xd5\xfb\x02\xdf24\"tV\xa7\xa8\xcc\x1e\x7f\x19Aw\x9b\xf2",  # noqa
        )


class TestSHA1HashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_sha1_hashing_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(SHA1HashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 20)
        self.assertEqual(
            output[0][1], b"@\x87\x98K\xa8u\xf9\x7f(\xd1\xa8=\x11\x9bcKT\xab\x1e\xe2"
        )


class TestMD5HashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_md5_hashing_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(MD5HashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 16)
        self.assertEqual(output[0][1], b"\xdb^R\x1f\xe0rP\xeb|o\x01\xda\xd9\xe4#\xfd")


class TestBLAKE2BHashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_blake2b_hashing_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(BLAKE2BHashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 64)
        self.assertEqual(
            output[0][1],
            b"\xe4\x17+\xc0\xf0#\xe4\x99T\x1eM\x11\x08\x82W\x1f%\x9a\xe6\xd5)\xc4\xeb\x8f\xafJqW\xb4_\x92T\x13\x83\xea\x8bb\xf4\n\xf7w\xc0\xd6\xe8\xf1J\xb7\rs\xc1\x02C\xe3L\x96a\x16\xf7\x91\xe6\x84Kb\xba",  # noqa
        )


class TestBLAKE2SHashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_blake2b_hashing_processor(self):
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(BLAKE2SHashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][1]), 32)
        self.assertEqual(
            output[0][1],
            b'T\xd5\xda\x91\x80\xae7,/\x94D\\\x82\xaf"*."\xfc\x94\xb0F\xfa\x10\xce<\xcbF\x04\xa5\xe2)',
        )


class TestUnknownAlgorithmHashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_unknown_algorithm_hashing_processor(self):
        with self.assertRaises(RuntimeError):
            self.set_up_processor(UnknownAlgorithmHashingProcessor)


# Context Hashing Test


class TestMD5ContextHashingProcessor(bspump.unittest.ProcessorTestCase):
    def test_md5_hashing_processor(self):
        events = [
            ({}, b"deadc0de"),
        ]

        self.set_up_processor(MD5ContextHashingProcessor)

        output = self.execute(events)
        self.assertEqual(len(output[0][0]["hash"]), 16)
        self.assertEqual(
            output[0][0]["hash"], b"\xdb^R\x1f\xe0rP\xeb|o\x01\xda\xd9\xe4#\xfd"
        )



import logging

import googleapiclient.discovery
from google.oauth2 import service_account

from ..abc.connection import Connection


L = logging.getLogger(__name__)


# https://developers.google.com/identity/protocols/OAuth2ServiceAccount


class GoogleDriveConnection(Connection):
    """
    GoogleDriveConnection allows BSPump application to use Google Drive API.
    It can be used by connectors to take care of authentication and generate
    Google Drive service - using API v3.
    This connection is synchronous and therefore all connectors using it
    are blocking.

    To aquire service_account_file, you will need to go to console.developers.google.com and create service account.
    Create a private key for your service account and download json private key file.
    The process is well documented here https://developers.google.com/identity/protocols/OAuth2ServiceAccount .
    Then it is necessary to delegate authority over a client account to created
    service account. You can do it in admin console: https://admin.google.com/
    For more details refere to the link above.

    For the delegation, G Suite domain administrator console is necessary.
    Free Google accounts can't be used as target of Google Drive connection.
    """

    ConfigDefaults = {
        "scopes": ["https://www.googleapis.com/auth/drive"],
        "service_account_file": "",
        "account_email": "",
        "output_queue_max_size": 100,
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self._scopes = self.Config["scopes"]
        self._service_account_file = self.Config["service_account_file"]
        self._account_email = self.Config["account_email"]

        self._output_queue_max_size = int(self.Config["output_queue_max_size"])

        self._credentials = None
        self._delegated_credentials = None

    def _get_service(self):
        if self._credentials is None:
            self._credentials = service_account.Credentials.from_service_account_file(
                self._service_account_file, scopes=self._scopes
            )

        if self._delegated_credentials is None:
            self._delegated_credentials = self._credentials.with_subject(
                self._account_email
            )

        svc = googleapiclient.discovery.build(
            "drive",
            "v3",
            credentials=self._delegated_credentials,
            cache_discovery=False,
        )
        return svc



import inspect
import sys

from bspump.asab import Configurable
from bspump import unittest


class TestConfigDefaults(unittest.TestCase):
    def test_default_value_is_not_none(self):
        to_inspect = []
        # For BSPump modules
        for _, module in inspect.getmembers(sys.modules["bspump"], inspect.ismodule):
            for _, klass in inspect.getmembers(module, inspect.isclass):
                if issubclass(klass, Configurable):
                    to_inspect.append(klass)

        # For BSPump classes
        for _, klass in inspect.getmembers(sys.modules["bspump"], inspect.isclass):
            if issubclass(klass, Configurable):
                to_inspect.append(klass)

        # Make unique
        to_inspect = list(set(to_inspect))

        for klass in to_inspect:
            for key, value in klass.ConfigDefaults.items():
                self.assertIsNotNone(value, f"None found for key {key} in {klass}")



import bspump.unittest
import bspump.common


class TestFlattenDictProcessor(bspump.unittest.ProcessorTestCase):
    def test_flatten_dict_processor(self):
        events = [
            (
                None,
                {
                    "person": {
                        "details": {
                            "first_name": "John",
                            "last_name": "Wick",
                        },
                        "address": {
                            "city": "Mill Neck",
                            "county": "USA",
                        },
                    }
                },
            ),
            (None, {}),
        ]

        self.set_up_processor(bspump.common.FlattenDictProcessor)

        output = self.execute(events)

        self.assertEqual(
            sorted([event for context, event in output], key=lambda x: len(x.keys())),
            [
                {
                    None: {},
                },
                {
                    "person.details.first_name": "John",
                    "person.details.last_name": "Wick",
                    "person.address.city": "Mill Neck",
                    "person.address.county": "USA",
                },
            ],
        )



import datetime
import pytz

from bspump.asab import Config

from ...abc import Expression
from ..value.valueexpr import VALUE


class DATETIME_FORMAT(Expression):
    """
    Returns date/time in human readable format.
    The date is created from `datetime`, which by default is current UTC time.

    Format example: "%Y-%m-%d %H:%M:%S"
    """

    Attributes = {
        "Format": ["*"],  # TODO: This ...
        "Value": ["*"],  # TODO: This ...
    }

    Category = "Date/Time"

    def __init__(self, app, *, arg_format, arg_with=None, arg_timezone=None):
        super().__init__(app)
        self.Value = arg_with if arg_with is not None else datetime.datetime.utcnow()
        if not isinstance(self.Value, Expression):
            self.Value = VALUE(app, value=self.Value)

        if not isinstance(arg_format, Expression):
            self.Format = VALUE(app, value=arg_format)
        else:
            self.Format = arg_format

        if arg_timezone is None:
            timezone_from_config = Config["declarations"]["timezone"]

            if len(timezone_from_config) == 0:
                self.Timezone = None

            else:
                self.Timezone = pytz.timezone(timezone_from_config)

        else:
            self.Timezone = pytz.timezone(arg_timezone)

    def __call__(self, context, event, *args, **kwargs):
        fmt = self.Format(context, event, *args, **kwargs)
        value = self.Value(context, event, *args, **kwargs)
        if isinstance(value, int) or isinstance(value, float):
            value = datetime.datetime.fromtimestamp(value, tz=self.Timezone)
        return value.strftime(fmt)



from ...abc import Expression


class EVENT(Expression):
    """
    The current event.

    Usage:
    ```
    !EVENT
    ``
    """

    Attributes = {}

    Category = "Value"

    def __init__(self, app, *, value):
        super().__init__(app)
        assert value == ""

    def __call__(self, context, event, *args, **kwargs):
        return event

    def get_outlet_type(self):
        return dict.__name__


class KWARGS(Expression):
    """
    The current kwargs.

    Usage:
    ```
    !KWARGS
    ``
    """

    Attributes = {}

    Category = "Value"

    def __init__(self, app, *, value):
        super().__init__(app)
        assert value == ""

    def __call__(self, context, event, *args, **kwargs):
        return kwargs


class KWARG(Expression):
    """
    The item from a kwargs.

    Usage:
    ```
    !KWARG argname
    ``
    """

    Attributes = {}
    Category = "Value"

    def __init__(self, app, *, value):
        super().__init__(app)
        self.ArgName = value

    def __call__(self, context, event, *args, **kwargs):
        return kwargs[self.ArgName]


class ARGS(Expression):
    Attributes = {}
    Category = "Value"

    def __init__(self, app, *, value):
        super().__init__(app)
        assert value == ""

    def __call__(self, context, event, *args, **kwargs):
        return args


class ARG(Expression):
    Attributes = {}
    Category = "Value"

    def __init__(self, app, *, value):
        super().__init__(app)
        assert value == ""
        self.ArgNumber = 0

    def __call__(self, context, event, *args, **kwargs):
        return args[self.ArgNumber]



class Singleton(type):

    """
    The `singleton pattern <https://en.wikipedia.org/wiki/Singleton_pattern>`_ is a software design pattern that restricts the instantiation of a class to one object.

    *Note*: The implementation idea is borrowed from "`Creating a singleton in Python <https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python>`_" question on StackOverflow.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    @classmethod
    def delete(cls, singleton_cls):
        """
        The method for an intentional removal of the singleton object.
        It shouldn't be used unless you really know what you are doing.

        One use case is a unit test, which removes an Application object after each iteration.
        """
        del cls._instances[singleton_cls]



try:
    import orjson
except ModuleNotFoundError:
    pass

# IMPORTANT: This module is obsolete, not supported and will be removed in a future

from ..abc.processor import Processor


class DictToJsonBytesParser(Processor):
    """
    DictToJsonBytesParser transforms a dictionary to JSON-string encoded in bytes.
    The encoding charset can be specified in the configuration in `encoding` field.
    """

    ConfigDefaults = {
        "encoding": "utf-8",
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)
        self.Encoding = self.Config["encoding"]

    def process(self, context, event):
        assert isinstance(event, dict)
        return orjson.dumps(event)



from unittest.mock import patch

import bspump.filter
import bspump.unittest


class TestTimeDriftFilter(bspump.unittest.ProcessorTestCase):
    @patch("asab.metrics.metrics.Counter.flush")
    @patch("asab.application.Application.time")
    def test_time_drift_filter(self, mocked_time, mocked_flush):
        mocked_time.return_value = 100
        events = [
            (None, {"@timestamp": 40}),
            (None, {"@timestamp": 39}),
            (None, {"no": "timestamp"}),
        ]

        self.set_up_processor(bspump.filter.TimeDriftFilter)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{"@timestamp": 40}])
        self.assertEqual(
            self.Pipeline.Processor.TimeDriftFilterCounter.rest_get()["Values"],
            {"event.in": 3, "event.out": 1, "event.drop": 1, "timestamp.error": 1},
        )
        # TimeDriftFilter + UnitTestSink + profiler for both
        self.assertEqual(mocked_flush.call_count, 4)



import time


class CookieSessionMixIn(object):

    """
    This mixin allows to store session id in browser Cookies
    """

    def set_cookie_name(self, cookie_name):
        self.CookieName = cookie_name

    def get_session_id_from_request(self, request):
        return self._get_cookie(request)

    def set_session_to_response(self, session, response):
        assert session.Id is not None
        self._set_cookie(response, session.Id, max_age=session.MaxAge)

    def _get_cookie(self, request):
        cookie = request.cookies.get(self.CookieName)
        return cookie

    def _set_cookie(self, response, cookie_data, max_age=None):
        params = {}  # dict(self._cookie_params)
        if max_age is not None:
            params["max_age"] = max_age
            params["expires"] = time.strftime(
                "%a, %d-%b-%Y %T GMT", time.gmtime(time.time() + max_age)
            )

        if not cookie_data:
            response.del_cookie(
                self.CookieName, domain=params["domain"], path=params["path"]
            )
        else:
            response.set_cookie(self.CookieName, cookie_data, **params)



import os
import shutil
import filecmp


def synchronize_dirs(target, source):
    """
    Synchronizes 'source' directory into 'target' directory.
    The 'source' directory remains unchanged.

    1. Recursively walk through the "source" directory and compare files with the "target".
    2. If a file exists in "source" but not in "target", copy it.
    3. If a file exists in both but has been modified in "source", copy it to overwrite the one in "target".
    4. If a directory or file exists in "target" but not in "source", remove it.
    """

    # Ensure target directory exists
    if not os.path.exists(target):
        os.makedirs(target)

    # Step 1: Recursively copy files from source to target
    for dirpath, dirnames, filenames in os.walk(source):
        # Compute relative path to the source base
        relpath = os.path.relpath(dirpath, source)
        target_dir = os.path.join(target, relpath)

        # Create directories in target if they don't exist
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)

        # Check files and synchronize
        for filename in filenames:
            source_file = os.path.join(dirpath, filename)
            target_file = os.path.join(target_dir, filename)

            # Copy if the file doesn't exist in target or if it's modified
            if not os.path.exists(target_file) or not filecmp.cmp(
                source_file, target_file, shallow=False
            ):
                shutil.copy2(source_file, target_file)

    # Step 2: Recursively delete files/folders in target that don't exist in source
    for dirpath, dirnames, filenames in os.walk(
        target, topdown=False
    ):  # topdown=False for depth-first
        relpath = os.path.relpath(dirpath, target)
        source_dir = os.path.join(source, relpath)

        # Check and remove files not in source
        for filename in filenames:
            target_file = os.path.join(dirpath, filename)
            source_file = os.path.join(source_dir, filename)

            if not os.path.exists(source_file):
                os.remove(target_file)

        # Check and remove directories not in source
        for dirname in dirnames:
            target_subdir = os.path.join(dirpath, dirname)
            source_subdir = os.path.join(source_dir, dirname)

            if not os.path.exists(source_subdir):
                shutil.rmtree(target_subdir)



import csv
import logging
import os

from ..abc.sink import Sink

#

L = logging.getLogger(__file__)


#


class FileCSVSink(Sink):
    """
    Description:

    ** Default Config**

    path : ''

    dialect : 'excel'

    delimiter : ','

    doublequote : True

    escapechar : ""

    lineterminator : os.linesep

    quotechar : '"'

    quoting : csv.QUOTE_MINIMAL

    skipinitialspace : False

    strict : False

    """

    ConfigDefaults = {
        "path": "",
        "dialect": "excel",
        "delimiter": ",",
        "doublequote": True,
        "escapechar": "",
        "lineterminator": os.linesep,
        "quotechar": '"',
        "quoting": csv.QUOTE_MINIMAL,  # 0 - 3 for [QUOTE_MINIMAL, QUOTE_ALL, QUOTE_NONNUMERIC, QUOTE_NONE]
        "skipinitialspace": False,
        "strict": False,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Dialect = csv.get_dialect(self.Config["dialect"])
        self._csv_writer = None

    def get_file_name(self, context, event):
        """
        Description: Override this method to gain control over output file name.

        **Parameters**

        context :

        event :

        :return: path of context and config

        |

        """
        # Here we are able to modify the sink behavior from outside using context.
        # If provided, the filename (and path) is taken from the context instead of the config
        return context.get("path", self.Config["path"])

    def writer(self, f, fieldnames):
        """
        Description:

        **Parameters**

        f :

        fieldnames : file
                Name of the file.

        :return: dialect and fieldnames

        |

        """
        kwargs = dict()

        kwargs["delimiter"] = self.Config.get("delimiter")
        kwargs["doublequote"] = bool(self.Config.get("doublequote"))

        escape_char = self.Config.get("escapechar")
        escape_char = None if escape_char == "" else escape_char
        kwargs["escapechar"] = escape_char

        kwargs["lineterminator"] = self.Config.get("lineterminator")
        kwargs["quotechar"] = self.Config.get("quotechar")
        kwargs["quoting"] = int(self.Config.get("quoting"))
        kwargs["skipinitialspace"] = bool(self.Config.get("skipinitialspace"))
        kwargs["strict"] = bool(self.Config.get("strict"))

        return csv.DictWriter(f, dialect=self.Dialect, fieldnames=fieldnames, **kwargs)

    def process(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event : any data type
                Information with timestamp.

        """
        if self._csv_writer is None:
            # Open CSV file if needed
            fieldnames = event.keys()
            fname = self.get_file_name(context, event)
            self.fd = open(fname, "w", newline="")
            self._csv_writer = self.writer(self.fd, fieldnames)
            self._csv_writer.writeheader()

        self._csv_writer.writerow(event)
        self.fd.flush()

    def rotate(self):
        """
        Description: Call this to close the currently open file.

        """
        del self._csv_writer
        self._csv_writer = None



from bspump.http.web.components.field import Field
from bspump.http.web.template_env import env


class TextField(Field):
    def inner_html(self, default="", readonly=False):
        template = env.get_template("text-field.html")
        return template.render(
            default=default,
            default_classes=self.default_classes,
            default_input_props=self.default_input_props,
        )



from ...abc import Expression

from ..value.eventexpr import EVENT
from ..value.eventexpr import KWARGS
from ..value.eventexpr import ARG
from ..value.valueexpr import VALUE
from ..utility.context import CONTEXT


class ITEM(Expression):
    """
    Get the item from a dictionary.

    There are two forms:

    1) Mapping form

    !ITEM
    with: !EVENT
    item: foo
    default: 0

    2) Scalar form

    !ITEM EVENT potatoes

    Scalar form has some limitations (e.g no default value) but it is more compact
    """

    Attributes = {
        "With": ["*"],  # TODO: This ...
        "Item": ["*"],  # TODO: This ...
        "Default": ["*"],  # TODO: This ...
    }

    def __init__(
        self, app, *, arg_with=None, arg_item=None, arg_default=None, value=None
    ):
        super().__init__(app)

        self.OutletType = "^"  # Inlet type is to be set based on the parent advice

        if value is not None:
            # Scalar value provided
            with_, item = value.split(" ", 2)

            with_ = with_.upper()
            if with_ == "EVENT":
                self.With = EVENT(app, value="")
            elif with_ == "CONTEXT":
                self.With = CONTEXT(app, value="")
            elif with_ == "KWARGS":
                self.With = KWARGS(app, value="")
            elif with_ == "ARG":
                self.With = ARG(app, value="")
            else:
                raise RuntimeError(
                    "Invalid item argument '{}' - must be EVENT, CONTEXT, KWARGS, ARG",
                    format(with_),
                )

            self.Item = VALUE(app, value=item)
            self.Default = VALUE(app, value=None)

        else:
            if isinstance(arg_with, Expression):
                self.With = arg_with
            else:
                self.With = VALUE(app, value=arg_with)

            if isinstance(arg_item, Expression):
                self.Item = arg_item
            else:
                self.Item = VALUE(app, value=arg_item)

            if isinstance(arg_default, Expression):
                self.Default = arg_default
            else:
                self.Default = VALUE(app, value=arg_default)

    def optimize(self):
        if isinstance(self.With, EVENT) and isinstance(self.Item, VALUE):
            return ITEM_optimized_EVENT_VALUE(self)

        elif isinstance(self.With, CONTEXT) and isinstance(self.Item, VALUE):
            if "." in self.Item.Value:
                return ITEM_optimized_CONTEXT_VALUE_NESTED(self)
            else:
                return ITEM_optimized_CONTEXT_VALUE(self)

        return None

    def __call__(self, context, event, *args, **kwargs):
        wth = self.With(context, event, *args, **kwargs)
        item = self.Item(context, event, *args, **kwargs)

        try:
            return wth[item]

        except (KeyError, IndexError):
            if self.Default is None:
                return None
            return self.Default(context, event, *args, **kwargs)


class ITEM_optimized_EVENT_VALUE(ITEM):
    def __init__(self, orig):
        super().__init__(orig.App)

        self.With = orig.With
        self.Item = orig.Item
        self.Default = orig.Default

        if self.Default is None:
            self.DefaultValue = None
        elif isinstance(self.Default, VALUE):
            self.DefaultValue = self.Default(None, None)
        else:
            raise NotImplementedError("Default: {}".format(self.Default))

        self.Key = self.Item.Value
        self.OutletType = orig.OutletType

    def get_outlet_type(self):
        return self.OutletType

    def set_outlet_type(self, outlet_type):
        self.OutletType = outlet_type

    def optimize(self):
        # This is to prevent re-optimising the class
        return None

    def __call__(self, context, event, *args, **kwargs):
        return event.get(self.Key, self.DefaultValue)


class ITEM_optimized_CONTEXT_VALUE(ITEM):
    def __init__(self, orig):
        super().__init__(orig.App)

        self.With = orig.With
        self.Item = orig.Item
        self.Default = orig.Default

        self.Key = self.Item.Value

        if self.Default is None:
            self.DefaultValue = None
        elif isinstance(self.Default, VALUE):
            self.DefaultValue = self.Default(None, None)
        else:
            raise NotImplementedError("Default: {}".format(self.Default))

        self.OutletType = orig.OutletType

    def get_outlet_type(self):
        return self.OutletType

    def set_outlet_type(self, outlet_type):
        self.OutletType = outlet_type

    def optimize(self):
        # This is to prevent re-optimising the class
        return None

    def __call__(self, context, event, *args, **kwargs):
        return context.get(self.Key, self.Default)


class ITEM_optimized_CONTEXT_VALUE_NESTED(ITEM):
    def __init__(self, orig):
        super().__init__(orig.App)

        self.With = orig.With
        self.Item = orig.Item
        self.Default = orig.Default

        if self.Default is None:
            self.DefaultValue = None
        elif isinstance(self.Default, VALUE):
            self.DefaultValue = self.Default(None, None)
        else:
            raise NotImplementedError("Default: {}".format(self.Default))

        # TODO: Replace with JSON pointer path
        self.KeyList = self.Item.Value.split(".")

        self.OutletType = orig.OutletType

    def get_outlet_type(self):
        return self.OutletType

    def set_outlet_type(self, outlet_type):
        self.OutletType = outlet_type

    def optimize(self):
        # This is to prevent re-optimising the class
        return None

    def __call__(self, context, event, *args, **kwargs):
        value = context
        try:
            for key in self.KeyList:
                if isinstance(value, list):
                    value = value[int(key)]
                else:
                    value = value[key]
        except (TypeError, KeyError, IndexError):
            return self.DefaultValue

        return value



import json
import logging

import numpy as np

import bspump.asab as asab
from ..abc.lookup import Lookup
from ..matrix.sessionmatrix import SessionMatrix

###

L = logging.getLogger(__name__)

###


class MatrixLookup(Lookup):
    """
    Numpy Lookup
    """

    ConfigDefaults = {
        "update_period": 5,
    }

    def __init__(
        self,
        app,
        matrix_id=None,
        dtype="float_",
        on_clock_update=False,
        id=None,
        config=None,
        lazy=False,
    ):
        super().__init__(app, id=id, config=config, lazy=lazy)
        self.Indexes = {}

        svc = app.get_service("bspump.PumpService")
        if matrix_id is None:
            s_id = self.Id + "Matrix"
            self.Matrix = SessionMatrix(app, dtype, id=s_id)
            svc.add_matrix(self.Matrix)
        else:
            self.Matrix = svc.locate_matrix(matrix_id)

        self.MatrixPubSub = None
        self.Timer = None

        self.Target = None

        if self.is_master():
            if on_clock_update:
                self.UpdatePeriod = float(self.Config["update_period"])
                self.Timer = asab.Timer(app, self._on_clock_tick, autorestart=True)
                self.Timer.start(self.UpdatePeriod)

            else:
                self.MatrixPubSub = self.Matrix.PubSub
                self.MatrixPubSub.subscribe("Matrix changed!", self._on_matrix_changed)

    async def _on_matrix_changed(self, message):
        self.update_indexes()

    async def _on_clock_tick(self):
        self.update_indexes()

    def update_indexes(self):
        for index in self.Indexes:
            self.Indexes[index].update(self.Matrix)

    def search(self, condition, target_column):
        """
        Default search, override if optimized with indexes
        """
        x = np.where(condition)
        if len(x[0]) == 0:
            return None

        return np.asscalar(self.Matrix.Array[x[0][0]][target_column])

    def serialize(self):
        serialized = {}
        serialized["Matrix"] = self.Matrix.serialize()
        serialized["Indexes"] = {}
        for index in self.Indexes:
            serialized["Indexes"][index] = self.Indexes[index].serialize()

        return (json.dumps(serialized)).encode("utf-8")

    def deserialize(self, data_json):
        data = json.loads(data_json.decode("utf-8"))
        self.Matrix.deserialize(data["Matrix"])
        indexes = data.get("Indexes", {})
        for index in indexes:
            self.Indexes[index].deserialize(indexes[index])

    def rest_get(self):
        rest = super().rest_get()
        rest["Indexes"] = {}
        for index in self.Indexes:
            rest["Indexes"][index] = self.Indexes[index].serialize()
        rest["Matrix"] = self.Matrix.serialize()
        return rest

    def create_index(self, index_class, *args, **kwarg):
        new_index = index_class(*args, **kwarg)
        index_id = new_index.Id
        self.Indexes[index_id] = new_index
        return new_index



from ..abc.processor import Processor
import random


class RandomEnricher(Processor):
    """
    `RandomEnricher` is mostly meant for testing.
    It adds to event only one field with the name from `Config['field']`, default is "id".
    There can be 2 options of usage:
    a) User provides `choice` as an array of values to choice randomly and assigned to field;
    b) The random integer between `Config['lower_bound']` and `Config['upper_bound']`.
    """

    ConfigDefaults = {
        "field": "enriched_field",
        "lower_bound": 0,
        "upper_bound": 1000,
    }

    def __init__(self, app, pipeline, choice=None, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Field = self.Config["field"]
        self.LowerBound = int(self.Config["lower_bound"])
        self.UpperBound = int(self.Config["upper_bound"])
        self.Choice = None
        if choice is not None:
            self.Choice = choice

    def generate_random(self, context, event):
        """
        Override this method to generate differently
        """
        if self.Choice is not None:
            event[self.Field] = random.choice(self.Choice)
        else:
            n = random.randint(self.LowerBound, self.UpperBound)
            event[self.Field] = n

    def process(self, context, event):
        self.generate_random(context, event)
        return event



import logging

import kazoo.retry
import kazoo.client
import kazoo.exceptions

#

L = logging.getLogger(
    __name__.rsplit(".", 1)[0]
)  # We want just "asab.zookeeper" in error messages

#


class KazooWrapper(object):
    def __init__(self, zkcnt, hosts):
        self.App = zkcnt.App
        self.ProactorService = zkcnt.ProactorService

        self.Client = kazoo.client.KazooClient(
            hosts=hosts,
            connection_retry=kazoo.retry.KazooRetry(
                max_tries=-1,  # Try to reconnect indefinetively
            ),
        )

        self.Client.add_listener(zkcnt._listener)

    async def _stop(self):
        ret = await self.ProactorService.execute(
            self.Client.stop,
        )
        return ret

    # read-only calls
    async def ensure_path(self, path):
        ret = await self.ProactorService.execute(self.Client.ensure_path, path)
        return ret

    async def exists(self, path):
        ret = await self.ProactorService.execute(self.Client.exists, path)
        return ret

    async def get_children(self, path):
        try:
            children = await self.ProactorService.execute(
                self.Client.get_children, path
            )
        except kazoo.exceptions.NoNodeError:
            # This is a silent error, it is indicated by None in the return
            return None
        return children

    async def get_data(self, path):
        try:
            data, stat = await self.ProactorService.execute(self.Client.get, path)
        except kazoo.exceptions.NoNodeError:
            # This is a silent error, it is indicated by None in the return
            return None
        return data

    # write methods
    async def set_data(self, path, data):
        try:
            ret = await self.ProactorService.execute(self.Client.set, path, data)
        except kazoo.exceptions.NoNodeError:
            L.warning(
                "Failed to write the data. Reason: Node '{}' does not exist.".format(
                    path
                )
            )
            return None
        return ret

    async def delete(self, path, version=-1, recursive=False):
        try:
            ret = await self.ProactorService.execute(
                self.Client.delete, path, version, recursive
            )
        except kazoo.exceptions.NoNodeError:
            L.warning(
                "Failed to delete node. Reason: Node '{}' does not exist.".format(path)
            )
            return None
        return ret

    async def create(
        self, path, value, sequence=False, ephemeral=False, makepath=False
    ):
        def do():
            return self.Client.create(
                path,
                value=value,
                ephemeral=ephemeral,
                sequence=sequence,
                makepath=makepath,
            )

        return await self.ProactorService.execute(do)

    # create a new node or update the existing one
    async def upsert(self, path, value):
        def do():
            try:
                return self.Client.create(path, value=value)
            except kazoo.exceptions.NodeExistsError:
                return self.Client.set(path, value)

        return await self.ProactorService.execute(do)



import sys
from unittest.mock import patch, call

import bspump.common
import bspump.unittest


class TestPrintSink(bspump.unittest.ProcessorTestCase):
    @patch("builtins.print")
    def test_print_sink(self, mocked_print):
        events = [
            (None, "Please, print this to me"),
        ]

        self.set_up_processor(bspump.common.PrintSink)

        output = self.execute(events)
        self.assertEqual(
            mocked_print.mock_calls[1],
            call("Please, print this to me", file=sys.stdout),
        )

        self.assertEqual([event for context, event in output], [])


class TestPPrintSink(bspump.unittest.ProcessorTestCase):
    @patch("pprint.PrettyPrinter.pprint")
    def test_pprint_sink(self, mocked_print):
        event = {"list": ["a", "b"], "dict": {"a": "b"}}
        events = [
            (None, event),
        ]

        self.set_up_processor(bspump.common.PPrintSink)

        output = self.execute(events)

        mocked_print.assert_called_with(event)

        self.assertEqual([event for context, event in output], [])


class TestPrintProcessor(bspump.unittest.ProcessorTestCase):
    @patch("builtins.print")
    def test_print_processor(self, mocked_print):
        event = "Please, print this to me"
        events = [
            (None, event),
        ]

        self.set_up_processor(bspump.common.PrintProcessor)

        output = self.execute(events)
        self.assertEqual(
            mocked_print.mock_calls[1],
            call("Please, print this to me", file=sys.stdout),
        )

        self.assertEqual([event for context, event in output], [event])


class TestPPrintProcessor(bspump.unittest.ProcessorTestCase):
    @patch("pprint.PrettyPrinter.pprint")
    def test_pprint_processor(self, mocked_print):
        event = {"list": ["a", "b"], "dict": {"a": "b"}}
        events = [
            (None, event),
        ]

        self.set_up_processor(bspump.common.PPrintProcessor)

        output = self.execute(events)

        mocked_print.assert_called_with(event)

        self.assertEqual([event for context, event in output], [event])


class TestPrintContextProcessor(bspump.unittest.ProcessorTestCase):
    @patch("builtins.print")
    def test_print_context_processor(self, mocked_print):
        context = {"context": "dict"}
        events = [
            (context, None),
        ]

        self.set_up_processor(bspump.common.PrintContextProcessor)

        output = self.execute(events)
        self.assertEqual(mocked_print.mock_calls[1], call(context, file=sys.stdout))

        self.assertEqual([event for context, event in output], [])


class TestPPrintContextProcessor(bspump.unittest.ProcessorTestCase):
    @patch("pprint.PrettyPrinter.pprint")
    def test_pprint_context_processor(self, mocked_print):
        context = {"context": "dict"}
        events = [
            (context, None),
        ]

        self.set_up_processor(bspump.common.PPrintContextProcessor)

        output = self.execute(events)

        mocked_print.assert_called_with(context)

        self.assertEqual([event for context, event in output], [])



import collections.abc
import logging

from ..abc.generator import Generator
from ..abc.source import TriggerSource

L = logging.getLogger(__name__)


class IteratorSource(TriggerSource):
    """
    Description:

    |

    """

    def __init__(
        self, app, pipeline, iterator: collections.abc.Iterator, id=None, config=None
    ):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Iterator = iterator

    async def cycle(self, *args, **kwags):
        """
        Description:

        |

        """
        for event in self.Iterator:
            await self.process(event)


class IteratorGenerator(Generator):
    """
    Description:

    |

    """

    async def generate(self, context, event, depth):
        """
        Description:

        |

        """
        for item in event:
            self.Pipeline.inject(context, item, depth)



import os

import bspump.declarative
import bspump.unittest


class TestDeclarativeRegex(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_regex_parse_01(self):
        decl = self.load("./test_regex_parse_01.yaml")

        res = decl({}, {})

        self.assertEqual(
            res,
            {
                "first": "foo",
                "second": "123",
                "third": "bar with postfix",
                "item1": "foo",
            },
        )

    def test_regex_parse_02(self):
        decl = self.load("./test_regex_parse_02.yaml")

        res = decl({}, {})
        self.assertEqual(
            res,
            {
                ".tmp14": "6620",
                ".tmp15": "2388",
                ".tmp16": "137.124453",
                "act": "ACCEPTED",
                "app": "https",
                "cs1": "FAILED",
                "cs2": "EXT-OWA-OK",
                "deviceExternalId": "TCPP-888",
                "dhost": "192.168.24.1",
                "dpt": 443,
                "dst": "192.168.24.1",
                "duser": "<NULL>",
                "in": 6620,
                "out": 2388,
                "proto": "TCP",
                "severity": "I",
                "shost": "52.125.141.18",
                "spt": 35822,
                "src": "52.125.141.18",
            },
        )



from .container import ZooKeeperContainer
from ..abc.module import Module


class Module(Module):
    def __init__(self, app):
        super().__init__(app)

        from .service import ZooKeeperService

        self.Service = ZooKeeperService(app, "asab.ZooKeeperService")


__all__ = [
    "ZooKeeperContainer",
    "Module",
]



import logging
from ..abc.source import TriggerSource

L = logging.getLogger(__name__)


class UnitTestSource(TriggerSource):
    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Input = []

    async def cycle(self, *args, **kwags):
        try:
            for context, event in self.Input:
                await self.process(event, context=context)
        except Exception as e:
            L.exception("During unit test")
            raise e



import asyncio
import logging

import pymysqlreplication

from ..abc.source import Source

#

L = logging.getLogger(__name__)

#


class MySQLBinaryLogSource(Source):
    ConfigDefaults = {
        "server_id": 1,
        "log_file": "",  # name of the first log file
        "log_pos": 4,
        "auto_position": False,  # If server supports GTID, scroll to latest replicated position
        "extract_events": """
			DeleteRowsEvent
			UpdateRowsEvent
			WriteRowsEvent
			""",
    }

    """
	Before you started using it, make sure that
	my.cnf configuration file contains following lines:

	[mysqld]
	server-id		 = 1
	log_bin			 = /var/log/mysql/mysql-bin.log
	expire_logs_days = 10
	max_binlog_size  = 100M
	binlog-format    = row #Very important if you want to receive write, update and delete row events

	Also don't forget to restart the server after changing the config file.

	By default, the it picks the last log file, but you can specify your own, you can find it in /var/log/mysql/ folder, so it
	will process all files, started from specified one.

	The source extracts events from MySQL binary log. By default it extracts only [DeleteRowsEvent, UpdateRowsEvent, WriteRowsEvent],
	but you can extend the list in configuration with [QueryEvent, RotateEvent,StopEvent,FormatDescriptionEvent,XidEvent,
		GtidEvent,BeginLoadQueryEvent,ExecuteLoadQueryEvent,TableMapEvent,HeartbeatLogEvent]

	Examples:
	UpdateRowsEvent event:
		{
			'@timestamp': 1551791161,
			'event_size': 47,
			'log_position': 1251,
			'read_bytes': 13,
			'rows': [
				{'after_values': {'data': 'World', 'data2': 'Hello', 'id': 1}},
				{'before_values': {'data': 'Hello', 'data2': 'World', 'id': 1}}
			],
			'type': 'UpdateRowsEvent'
		}

	DeleteRowsEvent event:
		{
			'@timestamp': 1551791168,
			'event_size': 29,
			'log_position': 1525,
			'read_bytes': 12,
			'rows': [{'values': {'data': 'World', 'data2': 'Hello', 'id': 1}}],
			'type': 'DeleteRowsEvent'
		}

	WriteRowsEvent event:
		{
			'@timestamp': 1551791153,
			'event_size': 29,
			'log_position': 959,
			'read_bytes': 12,
			'rows': [{'values': {'data': 'Hello', 'data2': 'World', 'id': 1}}],
			'type': 'WriteRowsEvent'
		}

	"""

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self._connection = pipeline.locate_connection(app, connection)
        self.MySQLSettings = {
            "host": self._connection._host,
            "port": self._connection._port,
            "user": self._connection._user,
            "passwd": self._connection._password,
        }

        self.ServerId = int(self.Config["server_id"])
        self.ProactorService = app.get_service("asab.ProactorService")

        self.ServerId = int(self.Config["server_id"])
        self.LogFile = self.Config["log_file"]
        self.LogPos = int(self.Config["log_pos"])
        self.AutoPosition = self.Config.getboolean("auto_position")
        self.ResumeStream = True

        if self.LogFile == "":
            self.LogFile = None
            self.LogPos = None
            self.ResumeStream = False

        extract_events = []
        lines = self.Config["extract_events"].split("\n")
        for line in lines:
            line_core = "".join(line.split())
            if len(line_core) == 0:
                continue
            extract_events.append(line_core)

        self.ExtractEvents = frozenset(extract_events)
        self.Loop = app.Loop
        self.Queue = asyncio.Queue(loop=self.Loop)
        self.Running = True
        self.Stream = None

    def stream_data(self):
        gtid = None
        if self.AutoPosition:
            cursor = self._connection.acquire_sync_cursor()
            cursor.execute("SELECT @@global.gtid_executed as gtid")
            row = cursor.fetchone()
            if row.get("gtid"):
                gtid = row.get("gtid")
                self.ResumeStream = True
                if self.LogFile is not None:
                    L.warning(
                        "Requested auto_position. Configuration log_file and log_pos is ignored."
                    )
            else:
                L.warning(
                    "MySQL server is not configured for GTID, auto_position cannot be used."
                )

        self.Stream = pymysqlreplication.BinLogStreamReader(
            connection_settings=self.MySQLSettings,
            server_id=self.ServerId,
            log_file=self.LogFile,
            log_pos=self.LogPos,
            resume_stream=self.ResumeStream,
            blocking=True,
            auto_position=gtid,
        )

        for binlogevent in self.Stream:
            if not self.Running:
                return False

            event = {}

            event_type = binlogevent.__class__.__name__

            if event_type not in self.ExtractEvents:
                continue

            # General info
            event["type"] = event_type
            event["@timestamp"] = binlogevent.timestamp
            event["log_position"] = binlogevent.packet.log_pos
            event["event_size"] = binlogevent.event_size
            event["read_bytes"] = binlogevent.packet.read_bytes

            if isinstance(binlogevent, pymysqlreplication.row_event.RowsEvent):
                event["table"] = binlogevent.table
                event["schema"] = binlogevent.schema
                event["primary_key"] = binlogevent.primary_key
                event["rows"] = binlogevent.rows

            # Specific info
            if event_type == "TableMapEvent":
                event["table_id"] = binlogevent.table_id
                event["schema"] = binlogevent.schema
                event["table"] = binlogevent.table
                event["columns"] = binlogevent.column_count

            if event_type == "GtidEvent":
                event["commit"] = binlogevent.commit_flag
                event["GTID_NEXT"] = binlogevent.gtid

            if event_type == "RotateEvent":
                event["position"] = binlogevent.position
                event["next_binlog_file"] = binlogevent.next_binlog

            if event_type == "XidEvent":
                event["transaction_id"] = binlogevent.xid

            if event_type == "HeartbeatLogEvent":
                event["current_binlog"] = binlogevent.ident

            if event_type == "QueryEvent":
                event["schema"] = binlogevent.schema
                event["execution_time"] = binlogevent.execution_time
                event["query"] = binlogevent.query

            if event_type == "BeginLoadQueryEvent":
                event["file_id"] = binlogevent.file_id
                event["block_data"] = binlogevent.block_data

            if event_type == "ExecuteLoadQueryEvent":
                event["slave_proxy_id"] = binlogevent.slave_proxy_id
                event["execution_time"] = binlogevent.execution_time
                event["schema_length"] = binlogevent.schema_length
                event["error_code"] = binlogevent.error_code
                event["status_vars_length"] = binlogevent.status_vars_length
                event["file_id"] = binlogevent.file_id
                event["start_pos"] = binlogevent.start_pos
                event["end_pos"] = binlogevent.end_pos
                event["dup_handling_flags"] = binlogevent.dup_handling_flags

            if event_type == "IntvarEvent":
                event["type"] = binlogevent.type
                event["value"] = binlogevent.value

            self.Queue.put_nowait(({}, event))
        return True

    async def main(self):
        worker = self.ProactorService.execute(self.stream_data)

        try:
            while True:
                await self.Pipeline.ready()
                context, event = await self.Queue.get()

                await self.process(event, context={})
                self.Queue.task_done()

        except asyncio.CancelledError:
            if self.Queue.qsize() > 0:
                self.Running = False
                L.warning(
                    "'{}' stopped with {} events in a queue".format(
                        self.locate_address(), self.Queue.qsize()
                    )
                )

        finally:
            self.Running = False
            worker.result()



import logging
from bspump.asab import Config
import bspump.asab as asab

from .service import ProactorService

#

L = logging.getLogger(__name__)

#

Config.add_defaults(
    {
        "asab:proactor": {
            "max_workers": "0",
            "default_executor": True,
        }
    }
)


class Module(asab.Module):
    """
    Proactor pattern based on loop.run_in_executor()

    https://en.wikipedia.org/wiki/Proactor_pattern
    """

    def __init__(self, app):
        super().__init__(app)
        self.service = ProactorService(app, "asab.ProactorService")



import logging
import asyncio
import random
from ..abc.source import TriggerSource


L = logging.getLogger(__name__)


class RandomSource(TriggerSource):
    """
    `RandomSource` is mostly meant for testing. It
    generates n (specified in `Config` as `number`, default is 1000) events per trigger fire.

    There can be 2 options of usage:
    a) User provides `choice` as an array of values to choose randomly;
    b) The random integer between `Config['lower_bound']` and `Config['upper_bound']`

    If `field` from `Config` is not an empty string, the random source generates dictionary
    event `{`field`: `generated value`}`. Otherwise it's a random number.
    """

    ConfigDefaults = {
        "field": "",
        "number": 1000,
        "lower_bound": 0,
        "upper_bound": 1000,
        "event_idle_time": 0.01,
        "events_till_idle": 10000,
    }

    def __init__(self, app, pipeline, choice=None, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Number = int(self.Config["number"])
        self.LowerBound = int(self.Config["lower_bound"])
        self.UpperBound = int(self.Config["upper_bound"])
        self.EventIdleTime = float(self.Config["event_idle_time"])
        self.EventsTillIdle = int(self.Config["events_till_idle"])
        self.EventCounter = 0
        self.Choice = None
        self.Field = None
        if choice is not None:
            self.Choice = choice

        if self.Config["field"] != "":
            self.Field = self.Config["field"]

    def generate_random(self):
        """
        Override this method to generate differently
        """
        if self.Choice is not None:
            if self.Field is not None:
                return {self.Field: random.choice(self.Choice)}

            return random.choice(self.Choice)
        else:
            if self.Field is not None:
                return {self.Field: random.randint(self.LowerBound, self.UpperBound)}

            return random.randint(self.LowerBound, self.UpperBound)

    async def cycle(self):
        for i in range(0, self.Number):
            event = self.generate_random()
            await self.process(event)
            if self.EventCounter >= self.EventsTillIdle:
                await asyncio.sleep(self.EventIdleTime)
                self.EventCounter = 0

            self.EventCounter += 1



from unittest.mock import MagicMock

import bspump.unittest
import bspump.common


class TestMappingKeysProcessor(bspump.unittest.ProcessorTestCase):
    def test_mapping_keys_processor(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
            (None, {"spam": "eggs"}),
        ]

        self.set_up_processor(bspump.common.MappingKeysProcessor)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [["foo"], ["fizz"], ["spam"]]
        )


class TestMappingValuesProcessor(bspump.unittest.ProcessorTestCase):
    def test_mapping_values_processor(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
            (None, {"spam": "eggs"}),
        ]

        self.set_up_processor(bspump.common.MappingValuesProcessor)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [["bar"], ["buzz"], ["eggs"]]
        )


class TestMappingItemsProcessor(bspump.unittest.ProcessorTestCase):
    def test_mapping_items_processor(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
            (None, {"spam": "eggs"}),
        ]

        self.set_up_processor(bspump.common.MappingItemsProcessor)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output],
            [[("foo", "bar")], [("fizz", "buzz")], [("spam", "eggs")]],
        )


class TestMappingKeysGenerator(bspump.unittest.ProcessorTestCase):
    def test_mapping_keys_generator(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
            (None, {"spam": "eggs"}),
        ]

        self.set_up_processor(bspump.common.MappingKeysGenerator)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], ["foo", "fizz", "spam"])


class TestMappingValuesGenerator(bspump.unittest.ProcessorTestCase):
    def test_mapping_values_generator(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
            (None, {"spam": "eggs"}),
        ]

        self.set_up_processor(bspump.common.MappingValuesGenerator)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], ["bar", "buzz", "eggs"])


class TestMappingItemsGenerator(bspump.unittest.ProcessorTestCase):
    def test_mapping_items_generator(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
            (None, {"spam": "eggs"}),
        ]

        self.set_up_processor(bspump.common.MappingItemsGenerator)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output],
            [("foo", "bar"), ("fizz", "buzz"), ("spam", "eggs")],
        )



import logging
import asyncio
import weakref
import functools
import typing


L = logging.getLogger(__name__)


class PubSub(object):
    """
    Object for delivering messages across the ASAB application.

    A message is a function or coroutine with specific `message_type` that can be published and subscribed at various places in the code.
    """

    def __init__(self, app):
        self.Subscribers = {}
        self.Loop = app.Loop

    def subscribe(self, message_type: str, callback: typing.Callable):
        """
        Set `callback` that will be called when `message_type` is received.

        Args:
                message_type: Message to be subscribed to. It should end with an exclamation mark `"!"`.
                callback: Function or coroutine that is called when the message is received. `message_type` is passed as the first argument to the callback.

        Examples:

        ```python
        class MyClass:
                def __init__(self, app):
                        app.PubSub.subscribe("Application.tick!", self.on_tick)

                def on_tick(self, message_type):
                        print(message_type)
        ```
        """

        # If subscribe is a bound method, do special treatment
        # https://stackoverflow.com/questions/53225/how-do-you-check-whether-a-python-method-is-bound-or-not
        if hasattr(callback, "__self__"):
            callback = weakref.WeakMethod(callback)
        else:
            callback = weakref.ref(callback)

        if message_type not in self.Subscribers:
            self.Subscribers[message_type] = [callback]
        else:
            self.Subscribers[message_type].append(callback)

    def subscribe_all(self, obj):
        """
        Find all methods decorated by `@asab.subscribe` on the object and subscribe for them.

        Examples:

        ```python
        class MyClass:
                def __init__(self, app):
                        app.PubSub.subscribe_all(self)

                @asab.subscribe("Application.tick!")
                async def on_tick(self, message_type):
                        print(message_type)

                @asab.subscribe("Application.exit!")
                def on_exit(self, message_type):
                        print(message_type)
        ```
        """
        for member_name in dir(obj):
            member = getattr(obj, member_name)
            message_types = getattr(
                member, "asab_pubsub_subscribe_to_message_types", None
            )
            if message_types is not None:
                for message_type in message_types:
                    self.subscribe(message_type, member)

    def unsubscribe(self, message_type, callback):
        """
        Remove `callback` from the subscribed `message_type`.

        When the subscription does not exist, warning is displayed.

        Examples:

        ```python
        class MyClass:
                def __init__(self, app):
                        app.PubSub.subscribe("Application.tick!", self.only_once)

                def only_once(self, message_type):
                        print("This message is displayed only once!")
                        app.PubSub.unsubscribe("Application.tick!", self.only_once)
        ```
        """
        callback_list = self.Subscribers.get(message_type)
        if callback_list is None:
            L.warning("Message type subscription '{}'' not found.".format(message_type))
            return

        remove_list = None

        for i in range(len(callback_list)):
            # Take an weakref entry in the callback list and references it
            c = callback_list[i]()

            # Check if a weak reference is working
            if c is None:  # a reference is lost, remove this entry
                if remove_list is None:
                    remove_list = list()
                remove_list.append(callback_list[i])
                continue

            if c == callback:
                callback_list.pop(i)
                break

        else:
            L.warning(
                "Subscriber '{}'' not found for the message type '{}'.".format(
                    message_type, callback
                )
            )

        if remove_list is not None:
            for callback_ref in remove_list:
                callback_list.remove(callback_ref)

        if len(callback_list) == 0:
            del self.Subscribers[message_type]

    def _callback_iter(self, message_type):
        callback_list = self.Subscribers.get(message_type)
        if callback_list is None:
            return

        remove_list = None

        for callback_ref in callback_list:
            callback = callback_ref()

            # Check if a weak reference is working
            if callback is None:  # a reference is lost
                if remove_list is None:
                    remove_list = list()
                remove_list.append(callback_ref)
                continue

            if asyncio.iscoroutinefunction(callback):
                callback = functools.partial(_deliver_async, self.Loop, callback)

            yield callback

        if remove_list is not None:
            for callback_ref in remove_list:
                callback_list.remove(callback_ref)

    def publish(self, message_type: str, *args, **kwargs):
        """
        Publish the message and notify the subscribers of an `message type`.

        `message_type` is passed as the first argument to the subscribed callback.

        Args:
                message_type: The emitted message.
                asynchronously (bool, optional): If `True`, `call_soon()` method will be used for the asynchronous delivery of the message. Defaults to `False`.

        Examples:

        ```python
        class MyApplication(asab.Application):
                async def initialize(self):
                        self.Count = 0
                        self.PubSub.subscribe("Fireworks.started!", self.on_fireworks)

                async def main(self):
                        for i in range(3):
                                self.Count += 1
                                self.PubSub.publish("Fireworks.started!", self.Count)
                                await asyncio.sleep(1)

                def on_fireworks(self, message_type, count):
                        print("boom " * count)
        ```
        """

        asynchronously = kwargs.pop("asynchronously", False)

        if asynchronously:
            for callback in self._callback_iter(message_type):
                self.Loop.call_soon(
                    functools.partial(callback, message_type, *args, **kwargs)
                )

        else:
            for callback in self._callback_iter(message_type):
                callback(message_type, *args, **kwargs)

    def publish_threadsafe(self, message_type: str, *args, **kwargs):
        """
        Publish the message and notify the subscribers of an `message type` safely form a different that main thread.

        `message_type` is passed as the first argument to the subscribed callback.

        Args:
                message_type: The emitted message.
                asynchronously (bool, optional): If `True`, `call_soon()` method will be used for the asynchronous delivery of the message. Defaults to `False`.
        """

        def in_main_thread():
            self.publish(message_type, *args, **kwargs)

        self.Loop.call_soon_threadsafe(in_main_thread)

    async def message(self, message_type: str) -> tuple:
        """
        Await specific message from a coroutine. It is a convenience method for the `Subscriber` object.

        Args:
                message_type: Message to be awaited.

        Returns:
                Triple (message_type, args, kwargs).

        Examples:

        ```python
        message_type, args, kwargs = await self.PubSub.message("Library.ready!")
        ```
        """
        subscriber = Subscriber(self, message_type)
        message_type, args, kwargs = await subscriber.message()
        return message_type, args, kwargs


class subscribe(object):
    """
    Decorator function that simplifies the process of subscription together with `PubSub.subscribe_all()` method.

    Examples:
    ```python
    class MyClass(object):
            def __init__(self, app):
                    app.PubSub.subscribe_all(self)

            @asab.subscribe("Application.tick!")
            async def on_tick(self, message_type):
                    print(message_type)

            @asab.subscribe("Application.exit!")
            def on_exit(self, message_type):
                    print(message_type)
    ```
    """

    def __init__(self, message_type):
        self.message_type = message_type

    def __call__(self, f):
        if getattr(f, "asab_pubsub_subscribe_to_message_types", None) is None:
            f.asab_pubsub_subscribe_to_message_types = [self.message_type]
        else:
            f.asab_pubsub_subscribe_to_message_types.append(self.message_type)
        return f


def _deliver_async_exited(task):
    try:
        task.result()
    except asyncio.CancelledError:
        pass


def _deliver_async(loop, callback, message_type, *args, **kwargs):
    task = asyncio.create_task(callback(message_type, *args, **kwargs))
    task.set_name("asab.PubSub.{}".format(message_type))
    task.add_done_callback(_deliver_async_exited)


class Subscriber(object):
    """
    Object for consuming PubSub messages in coroutines.

    It subscribes for various message types and consumes them.
    It is built on (first-in, first-out) basis.
    If `pubsub` argument is `None`, the initial subscription is skipped.

    Examples:

    The example of the subscriber object usage in async for statement:

    ```python
    async def my_coroutine(self):
            # Subscribe for two application events
            subscriber = asab.Subscriber(
                    self.PubSub,
                    "Application.tick!",
                    "Application.exit!"
            )
            async for message_type, args, kwargs in subscriber:
                    if message_type == "Application.exit!":
                            break;
                    print("Tick.")
    ```
    """

    def __init__(self, pubsub=None, *message_types):
        self._q = asyncio.Queue()
        self._subscriptions = []

        if pubsub is not None:
            for message_type in message_types:
                self.subscribe(pubsub, message_type)

    def subscribe(self, pubsub, message_type):
        """
        Subscribe for more message types. This method can be called many times with various `pubsub` objects.
        """
        pubsub.subscribe(message_type, self)
        self._subscriptions.append((pubsub, message_type))

    def __call__(self, message_type, *args, **kwargs):
        self._q.put_nowait((message_type, args, kwargs))

    def message(self):
        """
        Wait for a message asynchronously and return triple `(message_type, args, kwargs)`.

        Examples:

        ```python
        async def my_coroutine(app):
                # Subscribe for a two application events
                subscriber = asab.Subscriber(
                        app.PubSub,
                        "Application.tick!",
                        "Application.exit!"
                )
                while True:
                        message_type, args, kwargs = await subscriber.message()
                        if message_type == "Application.exit!":
                                break
                        print("Tick.")
        ```
        """
        return self._q.get()

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await self._q.get()
